"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import inspect
import shlex
import re
import sys
from pprint import pprint

command_pat = re.compile(r"[!@$?]?|[a-zA-Z0-9_-]+")
command_split_pat = re.compile(r'\s')


class ProxyNameSpace:
    def __init__(self, permanent_local, mylocal, *a, **ag):
        self.__global = permanent_local
        self.__local = mylocal

    def __getitem__(self, key):
        if key in self.__global:
            value = self.__global[key]
            self.__local[key] = value
            return value
        raise KeyError

    def __setitem__(self, key, value):
        self.__global[key] = value
        self.__local[key] = value

    def keys(self):
        return self.__global.keys()

    def items(self):
        return self.__global.items()

    def values(self):
        return self.__global.values()

    def __iter__(self):
        return self.__global.__iter__()


class BaseShell(object):
    intor_text = "Basic Shell By Md. Sabuj Sarker"
    prompt_text = "(sabuj): "

    def __init__(self, startup_commands=None, *args, **kwargs):
        self.__startup_commands = startup_commands
        self.symbol_to_method = {
            '@': 'py',
            '!': 'py',
            '?': 'help'
        }
        self.__loop_running = False
        self.__permanent_local_for_py = {}
        self.__last_set_value = None
        self.__setup_console_to_utf8()

    @staticmethod
    def __setup_console_to_utf8():
        sys.stdout = open(sys.stdout.fileno(), mode='w', encoding='utf8', buffering=1)

    def on_help(self, cmd=None):
        self.print("Help needed...")
        pass

    def on_empty(self):
        return True

    def close(self):
        self.__loop_running = False

    def on_print_last_set(self):
        self.print(self.__last_set_value)

    def on_py(self, *args):
        'write python code in one line'
        arg_0 = '' if not args else args[0]
        try:
            local = {}
            namespace = ProxyNameSpace(self.__permanent_local_for_py, local)
            line_buffer = []
            if arg_0 == "i":
                idx = 0
                while True:
                    idx += 1
                    line = self.input("%03d> " % idx)
                    if line == ":q":
                        "time to quit"
                        break
                    else:
                        line_buffer.append(line)
                        continue
            else:
                line_buffer.append(arg_0)
            src = "\n".join(line_buffer)
            self.print("Python Source:\n%s" % src)
            print("-------------")
            exec(src, globals(), namespace)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.print_error("Exception Type: %s" % exc_type.__name__)
            self.print_error("Exception: %s" % exc_value)

    def on_pyr(self, *args):
        "Recursive python - prompt another line after this one"
        'write python code in one line'

        idx = 0
        while True:
            idx += 1
            line = self.input("%04d >>> " % idx)
            line = line.strip()
            if line == ":q":
                "time to quit"
                break
            else:
                try:
                    local = {}
                    namespace = ProxyNameSpace(self.__permanent_local_for_py, local)
                    exec(line, globals(), namespace)
                    # for key, value in local.items():
                    #     self.print(key + ": " + str(value))
                except:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    self.print_error("Exception Type: %s" % exc_type.__name__)
                    self.print_error("Exception: %s" % exc_value)
                continue

    def on_p(self, *args):
        return self.on_py(args)

    def on_pi(self):
        return self.on_py("i")

    def print(self, *args, **kwargs):
        print(*args, *kwargs)

    def pprint(self, *args, **kwargs):
        pprint(*args, *kwargs)

    def print_error(self, *args, **kwargs):
        self.print(*args, **kwargs)

    def print_warning(self, *args, **kwargs):
        self.print(*args, **kwargs)

    def input(self, prompt):
        return input(prompt)

    def pre_loop(self):
        pass

    def on_shell(self, *cmd_args):
        self.pre_loop()

        self.__loop_running = True
        last_retcode = -1
        while self.__loop_running:
            inres = self.input(self.prompt_text).strip()
            if not inres:
                empty_ret = self.on_empty()
                if empty_ret:
                    continue
                else:
                    self.__loop_running = False
                    break
            else:
                shell_args = shlex.split(inres)
                last_retcode = self.run_command(shell_args)
                continue

        self.close()
        self.post_loop()
        return last_retcode

    def run_command(self, commands: list):
        cmd_args = commands
        if not cmd_args:
            self.print_error(f'No synamic command arg(s) provided to run.')
            return 1

        first_arg = cmd_args[0]
        rest_args = cmd_args[1:]

        if not command_pat.match(first_arg):
            self.print_error("Invalid first command format")
            return 1

        if first_arg in self.symbol_to_method:
            first_arg = self.symbol_to_method[first_arg]

        attr = getattr(self, "on_" + first_arg, None)

        if not attr:
            if first_arg in self.__permanent_local_for_py:
                self.print(self.__permanent_local_for_py[first_arg])
            else:
                self.print_error(f'Command not found {first_arg}')
            return 1

        if callable(attr):
            callable_attr = attr
            kwonlyargs = inspect.getfullargspec(callable_attr).kwonlyargs
            positional_args = inspect.getfullargspec(callable_attr).args
            no_of_positional_args = len(positional_args)
            self.print(kwonlyargs)

            if no_of_positional_args > 0 + 1:
                if not len(rest_args) >= no_of_positional_args:
                    self.print_error("Insufficient amount of shellargs provided for positional args.\n"
                                     f"Positional args: {positional_args}\n"
                                     f"shellargs: {rest_args}")
                    return 1
            return callable_attr(*rest_args)
        else:
            self.print_error(f"Invalid command provided: {first_arg}")
            return 1

    def on_exit(self, *args):
        'Exitting...'
        if len(args) > 0:
            retcode = args[0]
        else:
            retcode = ''
        self.print(f"Exiting Synamic Shell with retcode {retcode}")
        self.close()
        return retcode

    def post_loop(self):
        pass




class InitManager:
    def __init__(self, synamic):
        self.__synamic = synamic

    def init_site(self, cpath):
        assert cpath.is_dir
        dir_cpaths, file_cpaths = cpath.list_cpaths(respect_settings=False)
        if dir_cpaths or file_cpaths:
            print(f'The path {cpath.abs_path} is not empty. Init\'ing failed.')
            return False
        system_settings = self.__synamic.system_settings

        # gitignore
        cpath.join('.gitignore', is_file=True).\
            write_text(
            f'settings.private.syd\n'
            f'{system_settings["dirs.outputs.outputs"]}\n'
            f''
        )

        # settings
        cpath.join('settings.syd', is_file=True).\
            write_text('')

        # contents directory
        contents_cdir = cpath.join(system_settings['dirs.contents.contents'], is_file=False)
        contents_cdir.makedirs()
        contents_cdir.\
            join(system_settings['configs.dir_meta_file_name']).\
            write_text('type: post\n')
        contents_cdir.join('home.md', is_file=True).write_text(
'''
---
title: Welcome to Synamic
path: /
template: default.html
id: home
---

# Welcome to Synamic
This is your home page.

'''
        )

        # templates
        templates_cdir = cpath.join(system_settings['dirs.templates.templates'], is_file=False)
        templates_cdir.makedirs()
        templates_cdir.join('.gitkeep', is_file=True).make_file()
        templates_cdir.join('default.html', is_file=True).\
            write_text(
'''
<!doctype html>
<html>
    <head> <title> {{ content.title }} </title> </head>
    <body>
        {{ content.body.as_markup }}
    </body>
</html>
'''
        )

        # configs dir
        configs_cdir = cpath.join(system_settings['dirs.configs.configs'], is_file=False)
        configs_cdir.makedirs()
        configs_cdir.join('.gitkeep', is_file=True).make_file()

        # sites
        sites_cdir = cpath.join(system_settings['dirs.sites.sites'], is_file=False)
        sites_cdir.makedirs()
        sites_cdir.join('.gitkeep', is_file=True).make_file()

        # meta dirs
        metas_cdir = cpath.join(system_settings['dirs.metas.metas'], is_file=False)
        metas_cdir.makedirs()

        data_cdir = cpath.join(system_settings['dirs.metas.data'], is_file=False)
        markers_cdir = cpath.join(system_settings['dirs.metas.markers'], is_file=False)
        menus_cdir = cpath.join(system_settings['dirs.metas.menus'], is_file=False)
        models_cdir = cpath.join(system_settings['dirs.metas.models'], is_file=False)
        users_cdir = cpath.join(system_settings['dirs.metas.users'], is_file=False)

        # data
        data_cdir.makedirs()
        data_cdir.join('links.syd', is_file=True).write_text(
            '''
google: http://google.com
facebook: http://facebook.com
            '''
        )

        # markers
        markers_cdir.makedirs()
        markers_cdir.join('categories.syd', is_file=True).write_text(
            '''
title: Categories
type: hierarchical
marks:[
    {
        title: Programming
        id: programming
        description: Programming Category

        marks:[
            {
                title: C
                id: programming-c
                description: Programming C Category
            }
        ]
    }
    {
        title: Web Development
        id: web-development
        description: Web Development Category
        ext: ext-web-development
    }
    {
        title: Web Design
        id: web-design
        description: Web Design Category
        ext: ext-web-design
    }
    {
        title: Internet Marketing
        id: internet-marketing
        description: Internet Marketing Category
        ext: ext-internet-marketing
    }
    
    {
        title: Career
    }
]

            '''
        )
        markers_cdir.join('tags.syd', is_file=True).write_text(
            '''
title: Tags
type: multiple
slug: topic
marks:[
    {
        title: Marketing
        id: marketing
        description: Marketing Tag
    }
    {
        title: Design
        id: design
        description: Design Tag
    }
    {
        title: Programming
        id: programming
        description: Programming Tag
    }
    
    {
        title: Career
    }
]
            '''
        )
        markers_cdir.join('type.syd', is_file=True).write_text(
            '''
title: Type
type: single
is_public: 0
marks: [
    {
        title: Post
        id: post
        description: Post Type
    }
    {
        title: Page
        id: page
        description: Page Type
    }
    {
        title: Series
        id: series
    }
]

            '''
        )

        # menus
        menus_cdir.makedirs()
        menus_cdir.join('primary.syd', is_file=True).write_text(
            '''
menus [
    {
        title: Welcome Home
        link: url://id:home

        menus: [
            {
                title: Menu nested to another level.
                link: #
            }
        ]
    }
    {
        title: External Link
        link: https://www.sabuj.me
    }

]
            '''
        )

        # models
        models_cdir.makedirs()
        models_cdir.join('.gitkeep').make_file()

        # users
        users_cdir.makedirs()
        users_cdir.join('user1.syd').\
            write_text(
            '''
name: User One
title: User One's Profile

description ~ {
    User 1 is a very active user on this website.
}
            '''
        )

        return True



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import json
from synamic.core.standalones.functions.decorators import not_loaded, loaded
from synamic.core.contracts import DataContract
from synamic.exceptions import SynamicError, SynamicErrors, SynamicDataError


class JsonData(DataContract):
    def __init__(self, data_name, json_map):
        self.__name = data_name
        self.__json_map = json_map

    def get_data_name(self):
        return self.__name

    @property
    def origin(self):
        return self.__json_map

    def get(self, key, default=None):
        return self.__json_map.get(key, default=default)

    def __getitem__(self, item):
        return self.get(item, default=None)

    def __getattr__(self, key):
        return self.get(key, default=None)

    def __str__(self):
        return str(self.__json_map)

    def __repr__(self):
        return repr(self.__json_map)


class SydData(DataContract):
    def __init__(self, name, syd):
        self.__name = name
        self.__syd = syd

    def get_data_name(self):
        return self.__name

    @property
    def origin(self):
        return self.__syd

    def get(self, key, default=None):
        return self.__syd.get(key, default=default)

    def __getitem__(self, item):
        return self.get(item, None)

    def __getattr__(self, key):
        return self.get(key, None)

    def __str__(self):
        return str(self.__syd)

    def __repr__(self):
        return repr(self.__syd)


class DataService:
    """
    Currently supports: 
    - syd (xxx.syd)
    - json (xxx.json)

    Supports have been removed for: (As I want to keep things clean and before syd like stuffs with lot of features including all those data type and conversions were not in synamic)
    - yaml (xxx.yaml, xxx.yml)
    - ini (xxx.ini)
    - toml (xxx.toml)
    """
    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False

        data_dir = site.system_settings['dirs.metas.data']
        self.__data_cdir = site.path_tree.create_dir_cpath(data_dir)
        self.__available_extensions = ('.syd', '.json')

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    @loaded
    def make_data(self, data_name):
        syd_data_cfile = self.__data_cdir.join(f'{data_name}.syd', is_file=True)
        json_data_cfile = self.__data_cdir.join(f'{data_name}.json', is_file=True)
        res = None
        try:
            if syd_data_cfile.exists():
                syd = self.__site.object_manager.get_syd(syd_data_cfile)
                res = SydData(data_name, syd)
            elif json_data_cfile.exists():
                json_text = self.__site.object_manager.get_raw_text_data(json_data_cfile)
                json_obj = json.loads(json_text)
                res = JsonData(data_name, json_obj)
        except SynamicError as e:
            raise SynamicErrors(
                f'Data could not be made for data name {data_name}\n'
                f'One or both of the files tried: {syd_data_cfile.abs_path}, {json_data_cfile.abs_path}',
                e
            )

        if res is None:
            raise SynamicDataError(
                f'Data could not be made for data name {data_name}'
            )
        return res

    @loaded
    def get_data_names(self):
        names = []
        if self.__data_cdir.exists():
            data_file_cpaths = self.__data_cdir.list_files(checker=lambda cp: cp.basename.lower().endswith(self.__available_extensions))
            for file_cpath in data_file_cpaths:
                basename_wo_ext = file_cpath.basename_wo_ext
                names.append(basename_wo_ext)
        return names



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

from synamic.core.standalones.functions.decorators import not_loaded, loaded
from PIL import Image
from synamic.core.services.image_resizer.resized_image_content import ResizedImageContent


class ImageResizerService:
    __algo_map = {
        'nearest': Image.NEAREST,
        'bicubic': Image.BICUBIC,
        'bilinear': Image.BILINEAR,
        'lanczos': Image.LANCZOS
    }

    def __init__(self, site):
        self.__site = site
        self.__image_paths = set()  # image content is key and p
        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    @classmethod
    def __get_algorithm_flag(cls, algorithm):
        algorithm = algorithm.lower()
        if algorithm not in cls.__algo_map:
            algorithm_flag = Image.BICUBIC
        else:
            algorithm_flag = cls.__algo_map[algorithm]
        return algorithm_flag

    @loaded
    def resize(self, file_path, width, height, algorithm=''):
        # decide algorithm
        # algorithm_flag = self.__get_algorithm_flag(algorithm)
        # Algorithm is both unused here and in the img content class
        if type(file_path) is str:
            file_path = self.__site.path_tree.create_cpath(file_path, is_file=True)
        imgcnt = ResizedImageContent(self.__site, file_path, width, height)
        if imgcnt not in self.__image_paths:
            self.__image_paths.add(imgcnt)
            self.__site.add_auxiliary_content(imgcnt)
        return imgcnt



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

from synamic.core.standalones.functions.decorators import not_loaded, loaded
from .css import SCSS_CSSContent
from synamic.core.contracts import CDocType
from synamic.core.standalones.functions.sequence_ops import Sequence


class SASSProcessor:
    def __init__(self, site, processor_cpath):
        self.__site = site
        self.__processor_cpath = processor_cpath
        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    def get_generated_contents(self):
        content_objects = []
        content_service = self.__site.get_service('contents')
        if self.__processor_cpath.exists():
            sass_file_cpaths = self.__processor_cpath.list_files()
            for file_cpath in sass_file_cpaths:
                scss_basename = file_cpath.basename
                if file_cpath.extension.lower() in {'scss'}:
                    nourl_content_dirs_sw = self.__site.settings['configs.nourl_content_dirs_sw']
                    if not scss_basename.lower().startswith(nourl_content_dirs_sw):
                        # partial file in not not condition, ignore it.
                        curl = self.get_css_curl(file_cpath)
                        cdoctype = CDocType.GENERATED_TEXT_DOCUMENT
                        mimetype = 'text/css'
                        synthetic_fields = content_service.make_synthetic_cfields(
                            curl,
                            cdoctype,
                            mimetype,
                            cpath=None,
                            fields_map=None)
                        file_content = None

                        content_obj = SCSS_CSSContent(self.__site,
                                                      synthetic_fields,
                                                      file_content,
                                                      source_cpath=file_cpath)
                        content_objects.append(content_obj)
                else:
                    curl = self.get_static_file_curl(file_cpath)
                    cdoctype = CDocType.GENERATED_BINARY_DOCUMENT
                    mimetype = 'octet/stream'
                    synthetic_fields = content_service.make_synthetic_cfields(
                        curl,
                        cdoctype,
                        mimetype,
                        cpath=None,
                        fields_map=None)
                    file_content = None
                    content_obj = content_service.build_generated_content(
                        synthetic_fields,
                        curl, file_content,
                        cdoctype=CDocType.GENERATED_TEXT_DOCUMENT, mimetype=mimetype,
                        source_cpath=file_cpath
                    )
                    content_objects.append(content_obj)
        return content_objects

    def make_cpath(self, *path_comps, is_file=True):
        return self.__processor_cpath.join(*path_comps, is_file=is_file)

    def get_css_curl(self, scss_file_cpath):
        url_path_comps = Sequence.lstrip(scss_file_cpath.path_comps, self.__processor_cpath.parent_cpath.path_comps)
        assert len(url_path_comps) > 0

        scss_basename = url_path_comps[-1]
        parent_comps = url_path_comps[:-1]
        css_basename = scss_basename[:len(scss_basename) - len('.scss')] + '.css'

        css_url_path_comps = (*parent_comps, css_basename)
        curl = self.__site.synamic.router.make_url(
            self.__site, css_url_path_comps, for_cdoctype=CDocType.GENERATED_BINARY_DOCUMENT
        )
        return curl

    def get_static_file_curl(self, file_cpath):
        url_path_comps = Sequence.lstrip(file_cpath.path_comps, self.__processor_cpath.parent_cpath.path_comps)
        assert len(url_path_comps) > 0
        curl = self.__site.synamic.router.make_url(
            self.__site, url_path_comps, for_cdoctype=CDocType.GENERATED_BINARY_DOCUMENT
        )
        return curl



import abc


class BaseFsBackendContract(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def open(self, path, *args, **kwargs):
        """Return a file object"""

    @abc.abstractmethod
    def exists(self, path):
        """Checks existence of path"""

    @abc.abstractmethod
    def is_file(self, path):
        """Checks if it is a file"""

    @abc.abstractmethod
    def is_dir(self, path):
        """Checks if it is a dir"""

    @abc.abstractmethod
    def listdir(self, path):
        """Lists a directory"""

    @abc.abstractmethod
    def makedirs(self, path):
        """Makes directories recursively"""

    @abc.abstractmethod
    def getmtime(self, path):
        """Get modification? time"""

    @abc.abstractmethod
    def getctime(self, path):
        """Get last metadata change or creation time (on windows)"""

    @abc.abstractmethod
    def remove(self, path):
        """Removes the path"""



from .converters import _class_map, ConverterCallable, _default_types
"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


"""
1. Every module should have a .module.config.txt file at the root of it,
2. The module config file will have info about which data model it will follow.
3. Models will be stored inside 'models' of the site root folder.
4. Along with a default model per module's contents, a content file can specify a specific model with `_model` field.
"""


class TypeSystem:
    def __init__(self, site):
        self.__site = site
        self.__type_converters = {}
        self.__loaded = False

    @property
    def site(self):
        return self.__site

    def load(self):
        for name, converter_class in _class_map.items():
            self.add_converter(name, converter_class)
        self.__loaded = True

    def default_types(self):
        return _default_types

    def registered_types(self):
        return tuple(self.__type_converters.keys())

    def get_converter(self, type_name):
        converter = self.__type_converters.get(type_name, None)
        if converter is not None:
            return converter
        raise Exception("Converter with name `%s` not found" % type_name)

    def add_converter(self, type_name, converter):
        assert type_name not in self.__type_converters, "A converter with the type name already exists: `%s`" % type_name
        if issubclass(converter, ConverterCallable):
            _ = converter(self, type_name)
        else:
            _ = converter(self, type_name)  # they are the same for now as class must accept the type system
            #  and type name as the arguments
            assert callable(_), 'Provided class instance is not callable or __call__ is not defined on it: %s' % str(converter)
        converter = _
        self.__type_converters[type_name] = converter
        return converter



import re


def content_splitter(file_path, content_text):
    front_matter_sep = re.compile(r'^(?P<sep>-{3,})[ \t]*$')
    lines = content_text.splitlines()
    front_matter_lines = []
    body_lines = []

    fm_start_idx = None
    fm_end_idx = None
    sep = None
    idx = 0
    for line in lines:
        if fm_start_idx is None and fm_end_idx is None:
            front_matter_match = front_matter_sep.match(line)
            if line.strip() == '':
                # skip
                # before front matter
                idx += 1
                continue
            elif front_matter_match:
                sep = front_matter_match.group('sep')
                fm_start_idx = idx
            else:
                raise Exception('Invalid text before front matter section started. '
                                'Parsing error at line %d. File name: %s' % (idx + 1, file_path.relative_path))
        elif fm_start_idx is not None and fm_end_idx is None:
            # inside front matter.
            sep_end_text = line.strip()
            if sep_end_text == sep:
                fm_end_idx = idx
            else:
                front_matter_lines.append(line)
        elif fm_start_idx is not None and fm_end_idx is not None:
            # inside body
            body_lines.append(line)
        else:
            raise Exception('Parsing error at line %d. File name: %s' % (idx + 1, file_path.relative_path))
        idx += 1
    if fm_start_idx is None or fm_end_idx is None:
        raise Exception('Front matter section was not found. File name: %s' % file_path.relative_path)

    return '\n'.join(front_matter_lines), '\n'.join(body_lines)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import unittest

from synamic.core.standalones.functions import parse_front_matter


class TestFrontMatterParser(unittest.TestCase):
    def setUp(self):
        self.invalid_frontmatter1 = """---
        u:ttt
        hhhs-----
        """
        self.empty_frontmatter = """
        the content is here, no front matter around here.
        """
        self.valid_frontmatter = """----
name: My name
curl: somehow/curl
----
        """
        self.empty_text = ""

    def test_invalid(self):
        res = parse_front_matter(self.invalid_frontmatter1)
        self.assertTupleEqual(res, (None, None, None), "This should not be considered as a valid frontmattered text")

    def test_valid(self):
        res = parse_front_matter(self.valid_frontmatter)
        self.assertEqual(res[0], True)
        self.assertEqual(res[1], """name: My name
curl: somehow/curl""")
        self.assertEqual(res[2], """        """)

    def test_empty_frontmatter(self):
        res = parse_front_matter(self.empty_frontmatter)
        self.assertEqual(res[0], False, "No frontmatter must return False")
        self.assertEqual(res[2], self.empty_frontmatter)

    def test_empty_text(self):
        res = parse_front_matter(self.empty_text)
        self.assertTupleEqual(res[:2], (False, None), "the frontmatter is empty")
        self.assertEqual(res[2], "", "Body must be empty as the text is")



def is_sequence(target):
    return isinstance(target, (tuple, list))


def are_sequences(*args):
    assert len(args) > 0
    res = True
    for e in args:
        if not is_sequence(e):
            res = False
            break
    return res


class Sequence:
    @staticmethod
    def equals(target, what):
        assert are_sequences(target, what)
        if len(what) != len(target):
            return False
        else:
            equal = True
            for idx, e_target in enumerate(target):
                e_what = what[idx]
                if e_target != e_what:
                    equal = False
                    break
            return equal

    @staticmethod
    def startswith(target, what):
        assert are_sequences(target, what)
        if len(what) == 0 or len(target) == 0:
            return False
        elif len(what) > len(target):
            return False
        else:
            return Sequence.equals(target[:len(what)], what)

    @staticmethod
    def endswith(target, what):
        assert are_sequences(target, what)
        if len(what) == 0 or len(target) == 0:
            return False
        elif len(what) > len(target):
            return False
        else:
            return Sequence.equals(target[len(target) - len(what):], what)

    @staticmethod
    def lstrip(target, what):
        """Unlike string, just strips the first part"""
        assert are_sequences(target, what)
        res = target
        if Sequence.startswith(target, what):
            res = res[len(what):]
        return type(target)(res)

    @staticmethod
    def rstrip(target, what):
        """Unlike string, just strips the last part"""
        assert are_sequences(target, what)
        res = target
        if Sequence.endswith(target, what):
            res = res[:-len(what)]
        return type(target)(res)

    @staticmethod
    def strip(target, what):
        """Unlike string, just strips the first n last part"""
        assert are_sequences(target, what)
        res = target
        if Sequence.endswith(target, what):
            res = res[:-len(what)]
        if Sequence.startswith(target, what):
            res = res[len(what):]
        return type(target)(res)



if __name__ == '__main__':
    # TEST
    assert Sequence.equals([1, 2], list((1, 2)))
    assert Sequence.equals([1, 2], (1, 2))
    assert Sequence.equals([1, 2], [1, 2])
    assert Sequence.startswith((1, 2), [1, ])
    assert Sequence.endswith((1, 2), (2, ))
    assert Sequence.rstrip([1, 2, 3], [2, 3]) == [1], f'{Sequence.rstrip([1, 2, 3], [2, 3])}'
    assert Sequence.lstrip([1, 2, 3], [1, 2, 3]) == []
    assert Sequence.lstrip([1, 2, 3], [2, 2, 3]) == [1, 2, 3]
    assert Sequence.rstrip([1, 2, 3], [2, 2, 3]) == [1, 2, 3]
    print(f'Test positive!')



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


class DictUtils:
    @staticmethod
    def get_or_create_dict(d: dict, inner_dict_key):
        """
        Checks if another dictionary inside the dictionary exists or not. If does not exist it creates one. In either
        case it returns the dict 
        """
        if inner_dict_key not in d:
            d[inner_dict_key] = dict()
        return d[inner_dict_key]



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


"""
Site objects should be the public interface to access Synamic. Synamic objects should not be included as the 
public interfaces.
"""
import os
import sys
import re
import shutil
from collections import OrderedDict
from synamic.core.synamic.sites._site import _Site
from synamic.core.default_data._manager import DefaultDataManager
from synamic.core.standalones.functions.decorators import loaded, not_loaded
from synamic.exceptions import SynamicError, SynamicErrors


class Sites:
    def __init__(self, synamic, root_site_path):
        """
        :param root_site_path: Absolute path to the root site.
        """
        self.__synamic = synamic
        self.__root_site_path = root_site_path
        self.__default_data = None
        self.__sites_map = OrderedDict({
            #  site_id: site
        })  # must be ordered dict to keep serial of adding intact.
        self.__root_site = None
        self.__is_loaded = False

        # adding the root site
        _root_site_id = self.make_id('')
        assert _root_site_id.components == tuple()
        _root_site = self.make_site(_root_site_id, self.__root_site_path, parent_site=None, root_site=None)
        self.__root_site = _root_site
        self.add_site(_root_site)

    @property
    def is_loaded(self):
        return self.__is_loaded

    @property
    def ids(self):
        return tuple(self.__sites_map.keys())

    def make_id(self, comps):
        return _SiteId(self.__synamic, comps)

    def get_id_sep(self):
        return self.__synamic.system_settings['site_id_sep']

    def make_site(self, site_id, site_root_path_abs, parent_site=None, root_site=None):
        """Only this method should know how to make a site - no direct _Site class instantiation"""
        site_id = self.make_id(site_id)
        return _Site(self.__synamic, site_id, site_root_path_abs, parent_site=parent_site, root_site=root_site)

    def add_site(self, site):
        assert isinstance(site, _Site)
        self.__sites_map[site.id] = site

    def get_abs_path_by_id(self, site_id):
        """A site might not be located under the 'sites' dir, instead it can be configured in configs.
        Setting such custom path is not implemented yet, but will be in future."""
        site_id = self.make_id(site_id)
        return self.get_by_id(site_id).abs_path

    @not_loaded
    def load(self):
        assert os.path.exists(self.__root_site_path)
        # default configs
        self.__default_data = DefaultDataManager()

        # list all the site id components paths
        # this implements the default sites dire mechanism - externally located site is not implemented yet.
        children_site_comps_ids = []
        _subsites_dirname = self.__synamic.system_settings['configs.subsites_dir']
        self.__list_site_paths(children_site_comps_ids, tuple(), '')
        __sites_id_comps = tuple(children_site_comps_ids)
        for site_id_comps in __sites_id_comps:
            site_id = self.make_id(site_id_comps)
            parent_site_id = site_id.parent_id
            assert parent_site_id is not None
            parent_site = self.get_by_id(parent_site_id)
            root_site = self.__root_site
            site_root_abs_path = os.path.join(self.__synamic.abs_root_path,
                                              *self.__get_real_site_path_comps(site_id.components))
            site = self.make_site(site_id, site_root_abs_path, parent_site=parent_site, root_site=root_site)
            self.add_site(site)
        # load all the sites.
        for site in self.__sites_map.values():
            site.load()
        self.__is_loaded = True
        return self

    @property
    def root_site(self):
        return self.__root_site

    @property
    def root_site_path(self):
        return self.__root_site_path

    def get_by_id(self, site_id):
        site_id = self.make_id(site_id)
        if site_id in self.__sites_map:
            return self.__sites_map[site_id]
        raise KeyError('Site with id %s not found' % site_id)

    def __get_real_site_path_comps(self, site_virtual_comps: tuple):
        """Adds `sites` in between"""
        assert not isinstance(site_virtual_comps, str)
        subsites_dirname = self.__synamic.system_settings['configs.subsites_dir']
        real_comps = []
        for i in range(len(site_virtual_comps)):
            assert site_virtual_comps[i] != ''
            real_comps.append(subsites_dirname)
            real_comps.append(site_virtual_comps[i])
        # print('Real calc: of (%s)' % str(site_virtual_comps), real_comps)
        return tuple(real_comps)

    def __list_site_paths(self, result: list, v_path_comps_2_parent: tuple, start_with=''):
        virtual_site_comps = [*v_path_comps_2_parent]
        virtual_site_comps.append(start_with) if start_with != '' else None
        real_site_comps = self.__get_real_site_path_comps(tuple(virtual_site_comps))
        real_site_comps += (self.__synamic.system_settings['configs.subsites_dir'], )
        subsites_dir_abs = os.path.join(self.__root_site_path, *real_site_comps)
        if os.path.exists(subsites_dir_abs):
            subsite_ids = []
            for site_id in os.listdir(subsites_dir_abs):
                if os.path.isdir(os.path.join(subsites_dir_abs, site_id)):
                    subsite_ids.append(site_id)
            if subsite_ids:
                for subsite_id in subsite_ids:
                    site_id_comps = [*v_path_comps_2_parent]
                    if start_with != '':
                        site_id_comps.append(start_with)
                    site_id_comps = tuple(site_id_comps)
                    next_site_id_comps = (*site_id_comps, subsite_id)
                    result.append(next_site_id_comps)
                    self.__list_site_paths(
                        result,
                        site_id_comps,
                        subsite_id
                    )

    @loaded
    def build(self):
        try:
            # clean output directory

            output_dir = self.__synamic.system_settings['dirs.outputs.outputs']
            output_cdir = self.__synamic.path_tree.create_dir_cpath(output_dir)
            if not output_cdir.exists():
                output_cdir.makedirs()
            output_abs_path = output_cdir.abs_path
            except_root_paths = ('.git', '.gitignore', '.gitattributes')
            print(f'Removing paths from {output_abs_path} with some exceptions to {except_root_paths}')
            for o_basename in os.listdir(output_abs_path):
                full_path = os.path.join(output_abs_path, o_basename)
                if o_basename not in except_root_paths and not o_basename.startswith('.'):
                    print(f'Removing {full_path}')
                    if os.path.isfile(full_path):
                        os.remove(full_path)
                    else:
                        shutil.rmtree(full_path)

            # build sites
            build_succeeded = True
            for site_id, site in self.__sites_map.items():
                print(f'>>> Building Site: {site_id}\n\n')
                build_succeeded = site.object_manager.build()
                if not build_succeeded:
                    build_succeeded = False
                    break
        except SynamicError as e:
            e = SynamicErrors(
                f'Error building sites:',
                e
            )
            print(e, file=sys.stderr)
            build_succeeded = False
        return build_succeeded

    @loaded
    def upload(self):
        return self.__synamic.upload_manager.get_uploader('firebase').upload()


class _SiteId:
    __path_pat_split_by = re.compile(r'[\\/]')
    __ws_pat = re.compile(r'\s', re.MULTILINE)

    def __init__(self, synamic, comps):
        self.__synamic = synamic
        site_id_sep = self.__synamic.system_settings['configs.site_id_sep']
        _bk_comps = comps
        # start processing
        if isinstance(comps, self.__class__):
            self.__id_components = comps.components
        else:
            # normalizing to a list for easy iteration
            if isinstance(comps, (tuple, list)):
                _ = []
                for c in comps:
                    assert isinstance(c, str)
                    _.extend(c.split(site_id_sep))
                id_components = _
            else:
                assert isinstance(comps, str)
                id_components = comps.split(site_id_sep)

            # hunting down anything with / or \ and split at that position
            _ = []
            for c in id_components:
                _.extend(self.__path_pat_split_by.split(c))
            id_components = _

            # filter out any empty string
            _ = []
            for c in id_components:
                if c.strip() == '':
                    continue
                _.append(c)
            id_components = _

            # validating for invalid character
            for c in id_components:
                if self.__ws_pat.search(c):
                    raise Exception('Site id cannot contain white space characters: %s' % _bk_comps)

            # instance variable initialization
            self.__id_components = tuple(id_components)

    @property
    def components(self) -> tuple:
        return self.__id_components

    @property
    def path_as_ids(self):
        ids = []
        for c in self.components:
            ids.append(self.__class__(self.__synamic, c))
        return tuple(ids)

    @property
    def parent_components(self):
        if len(self.components) == 0:
            return None
        return self.components[:-1]

    @property
    def parent_id(self):
        parent_comps = self.parent_components
        if parent_comps is None:
            return None
        return self.__class__(self.__synamic, parent_comps)

    @property
    def as_string(self):
        site_id_sep = self.__synamic.system_settings['configs.site_id_sep']
        return site_id_sep.join(self.components)

    @property
    def is_root(self):
        return len(self.components) == 0

    def __len__(self):
        return len(self.components)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.components == other.components

    def __hash__(self):
        return hash(self.components)

    def __str__(self):
        return self.as_string

    def __repr__(self):
        return repr(self.__str__())



import os
from synamic.core.synamic import Synamic
from aiohttp import web
from .synamic_handler import synamic_handler

app = web.Application()
app.router.add_route('GET', '/{tail:.*}', synamic_handler)


def serve(root_dir: str, scheme='http', host='localhost', port='8080'):
    assert os.path.exists(root_dir)

    # bootstrap(site_root)
    synamic = Synamic(root_dir)
    # make site address
    port_str = f':{port}'
    site_address = scheme + "://" + host + port_str

    synamic.set_dev_params(site_address=site_address)

    synamic.load()
    #  synamic.sites.load()
    app.synamic = synamic
    web.run_app(app, host=host, port=port)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


from synamic.exceptions import CircularDependency


def create_dep_list(mods: dict):
    # sanity check : check that dict key and the mod name referred to by that name is the same.
    # Only the framework programmer can make this mistake!
    for name, mod in mods.items():
        if name != mod.name:
            raise Exception(
                "Fatal error - key name and mod name against that key must be the same!"
                "\nThe key name was `%s` and the mod name is `%s`" % (
                name, mod.name))

    # circular dependency checker - direct and non-direct
    for name, mod in mods.items():
        if len(mod.dependencies) == 0:
            continue

        _d_list = list(mod.dependencies.copy())  # initializing for first while loop
        _mod = mod

        # for detecting dependency chain
        dep_mod_chain = []

        while _mod is not None:
            dep_mod_chain.append(_mod)

            if name in _mod.dependencies:
                dep_mod_chain.append(mods[_mod.dependencies[_mod.dependencies.index(name)]])
                # chain representation:
                _str = ""
                for __mod in dep_mod_chain:
                    _str += "%s : %s\n" % (__mod.name, __mod.dependencies)
                raise CircularDependency("Case 1: circulur dependency. Dependency chain: \n%s" % _str)
            else:

                _d_list.extend(list(_mod.dependencies))

                if _d_list:
                    _mod = mods[_d_list[-1]]
                    del _d_list[-1]
                else:
                    _mod = None
                    # print("In the end of first while loop check done")
    # print("Dependency check done")
    my_mods = mods.copy()
    # zero dep modules
    done_mods = {}
    undone_mods = {}
    zero_dep_mods = {}
    dep_list = []

    # zero dep mod filtering
    for name, mod in my_mods.items():
        if len(mod.dependencies) == 0:
            zero_dep_mods[name] = mod
            dep_list.append(name)
        else:
            undone_mods[name] = mod

    if len(zero_dep_mods) == 0:
        raise Exception("At least one Zero Dependency Modules must be present")

    done_mods.update(zero_dep_mods)

    # dependency list building

    while len(undone_mods) > 0:
        for name, mod in undone_mods.copy().items():
            all_done = True
            for d_name in mod.dependencies:
                if d_name not in done_mods:
                    all_done = False
                    break
            if all_done:
                del undone_mods[name]
                done_mods[name] = mod
                dep_list.append(name)
    return dep_list



from synamic.core.services.marker.marker import Marker
from .marker import _Mark


class MarkerService:
    def __init__(self, site):
        self.__site = site

        self.__is_loaded = False

    def load(self):
        self.__is_loaded = True

    def make_marker(self, marker_id):
        markers_path = self.__site.system_settings['dirs.metas.markers']
        syd = self.__site.object_manager.get_syd(markers_path + '/' + marker_id + '.syd')
        marker_type = syd['type']
        marker_title = syd.get('title', marker_type)
        marker_mark_maps = syd['marks']
        marker = Marker(self.__site, marker_id, marker_type, marker_title, marker_mark_maps, syd)
        return marker

    def get_marker_ids(self):
        path_tree = self.__site.get_service('path_tree')
        markers_path = self.__site.system_settings['dirs.metas.markers']
        markers_cdir = path_tree.create_dir_cpath(markers_path)

        _ = []
        if markers_cdir.exists():
            marker_cpaths = markers_cdir.list_files(checker=lambda cp: cp.basename.endswith('.syd'))
            for cp in marker_cpaths:
                _.append(cp.basename[:-len('.syd')])
        ids = tuple(_)
        return ids

    def is_type_mark(self, other):
        return isinstance(other, _Mark)



from synamic.core.synamic.router.url import ContentUrl
from synamic.core.contracts import CDocType


class RouterService:
    def __init__(self, synamic):
        self.__synamic = synamic
        self.__is_loaded = False

    def load(self):
        self.__is_loaded = True

    def get_content(self, url_str: str):
        site_id, path_components, special_components = ContentUrl.parse_requested_url(self.__synamic, url_str)
        try:
            site = self.__synamic.sites.get_by_id(site_id)
        except KeyError:
            site = None

        if site is None:
            content = None
        else:
            # step 1: search for static/binary file in file system with the path components : TODO: do for static.
            # step 2 if 1 fails: search for non-static content and in this case the url is already cached.
            cdoctype = CDocType.UNSPECIFIED
            curl = self.make_url(site, path_components, cdoctype)
            if special_components:
                print(f'Special Components: {special_components}')
                pagination_url_comp = self.__synamic.system_settings['pagination_url_comp']
                mark_url_comp = self.__synamic.system_settings['mark_url_comp']
                user_url_comp = self.__synamic.system_settings['user_url_comp']

                if special_components[0] == user_url_comp:
                    if len(special_components) < 2:
                        return None
                    else:
                        user_id = special_components[1]
                        user = site.object_manager.get_user(user_id)
                        if user is None:
                            return None
                        else:
                            return user.content
                elif special_components[0] == mark_url_comp:
                    if len(special_components) < 3:
                        return None
                    else:
                        marker_id_or_slug = special_components[1]
                        mark_id = special_components[2]
                        marker = site.object_manager.get_marker(marker_id_or_slug, default=None)
                        if marker is None:
                            marker = site.object_manager.get_marker_by_slug(marker_id_or_slug)
                        if marker is None:
                            return None
                        mark = marker.get_mark_by_id(mark_id)
                        if mark is None:
                            return None
                        else:
                            return mark.content
                elif special_components[0] == pagination_url_comp:
                    if len(special_components) < 2:
                        return None
                    else:
                        page_no = special_components[1]
                        if page_no.isdigit():
                            page_no = int(page_no)
                        else:
                            return None

                        root_content = self.get_content_by_url(site, curl)
                        if root_content is None:
                            return None
                        else:
                            sub_page = root_content.pagination.get_sub_page(page_no)
                            if sub_page is None:
                                return None
                            else:
                                return sub_page.host_content
            content = self.get_content_by_url(site, curl)
        return content

    def get_content_by_url(self, site, curl):
        """Forgiving function that returns None"""
        print(f'Querying {curl.path_components} - doctype {curl.for_cdoctype} on site {site.id}')
        # try for text doc
        marked_content = site.object_manager.get_marked_content_by_curl(curl)
        if marked_content is not None:
            return marked_content

        pre_processed_content = site.object_manager.get_pre_processed_content_by_curl(curl)
        if pre_processed_content is not None:
            return pre_processed_content
        else:
            # TODO: fix bug: a.txt /a.txt/ and /a.txt work the same - /a.txt/ is most weird
            contents_dir = self.__synamic.system_settings['dirs.contents.contents']
            contents_dir_cpath = site.path_tree.create_dir_cpath(contents_dir)

            curl = curl.clone(CDocType.GENERATED_BINARY_DOCUMENT)

            fs_path = curl.to_file_system_path
            fs_path = fs_path.rstrip('/\\')
            file_cpath = contents_dir_cpath.join(fs_path, is_file=True)

            if file_cpath.exists():
                return site.object_manager.get_binary_content(file_cpath)
            else:
                return None

    @classmethod
    def make_url(cls, site, url_path_comps, for_cdoctype=CDocType.UNSPECIFIED):
        return ContentUrl(site, url_path_comps, for_cdoctype=for_cdoctype)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

from markupsafe import Markup
from synamic.core.standalones.functions.md import render_content_markdown


class Html:
    def __init__(self, html_str):
        self.__html_str = html_str
        self.__markup_safe = Markup(html_str)
        self.__plain_text = None

    @property
    def as_plain_text(self):
        if self.__plain_text is None:
            self.__plain_text = self.__markup_safe.striptags()
        return self.__plain_text

    def get_summary(self, limit=None):
        if limit is not None:
            return self.as_plain_text[:limit]
        else:
            return self.as_plain_text

    @property
    def as_str(self):
        return self.__html_str

    @property
    def as_markup(self):
        """Returns markup object from markupsafe"""
        return self.__markup_safe

    def __str__(self):
        return self.__html_str

    def __repr__(self):
        return repr(self.__str__())


class Markdown:
    def __init__(self, site, md_str, value_pack=None):
        self.__site = site
        self.__md_str = md_str
        self.__value_pack = value_pack
        self.__html_obj = None
        self.__plain_text = None
        self.__rendered_text = None

    @property
    def rendered_markdown(self):
        if self.__rendered_text is None:
            self.__rendered_text = render_content_markdown(self.__site, self.__md_str, value_pack=self.__value_pack)
        return self.__rendered_text

    @property
    def as_html(self):
        rendered_txt = self.rendered_markdown
        if self.__html_obj is None:
            self.__html_obj = Html(rendered_txt)
        return self.__html_obj

    @property
    def as_markup(self):
        return self.as_html.as_markup

    @property
    def as_plain_text(self):
        return self.as_html.as_plain_text

    def get_summary(self, limit=None):
        if limit is not None:
            return self.as_plain_text[:limit]
        else:
            return self.as_plain_text

    @property
    def as_str(self):
        return self.__md_str

    def __str__(self):
        return self.__md_str

    def __repr__(self):
        return repr(self.__str__())



import abc


class DataContract(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_data_name(self):
        pass

    @property
    @abc.abstractmethod
    def origin(self):
        """Returns original data object"""

    @abc.abstractmethod
    def get(self, key, default=None):
        """Given a key returns the data"""

    @abc.abstractmethod
    def __getitem__(self, item):
        """Uses get()"""

    @abc.abstractmethod
    def __getattr__(self, key):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def __repr__(self):
        pass



import sys
import os
import re
import subprocess
import datetime
ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


class FireBaseUploader:
    def __init__(self, synamic):
        self.__synamic = synamic

    def upload(self):
        path_tree = self.__synamic.path_tree
        output_dir = self.__synamic.system_settings['dirs.outputs.outputs']
        output_cdir = path_tree.create_dir_cpath(output_dir)
        extras_dir = self.__synamic.system_settings['dirs.configs.configs']
        extras_cdir = path_tree.create_dir_cpath(extras_dir)
        firebase_json_cfile = extras_cdir.join('firebase/firebase.json', is_file=True)
        firebase_rc_cfile = extras_cdir.join('firebase/.firebaserc', is_file=True)

        # checking for firebase files
        if not firebase_json_cfile.exists() or not firebase_rc_cfile.exists():
            print(f'Error: firebase.json and .firebaserc file must exists in extras: '
                  f'{firebase_json_cfile.abs_path} {firebase_rc_cfile.abs_path}', file=sys.stderr)
            return False
        root_site_settings = self.__synamic.sites.root_site.settings

        # firebase project name
        firebase_projce_name = os.getenv('FIREBASE_PROJECT_NAME', None)
        if firebase_projce_name is None:
            firebase_projce_name = root_site_settings.get('firebase.FIREBASE_PROJECT_NAME', None)
        if firebase_projce_name is None:
            print(f'Firebase project_name is not specified in environment, neither in site settings under firebase key', file=sys.stderr)
            return False

        # firebase ci token
        firebase_ci_token = os.getenv('FIREBASE_CI_TOKEN', None)
        if firebase_ci_token is None:
            firebase_ci_token = root_site_settings.get('firebase.FIREBASE_CI_TOKEN', None)

        if firebase_ci_token is None:
            print(f'FIREBASE_CI_TOKEN not found in environment, neither in settings', file=sys.stderr)
            return False

        # build first
        print(f'Building the sites before uploading')
        build_succeeded = self.__synamic.sites.build()
        if not build_succeeded:
            print(f'Building failed. Not uploading. See error log.', file=sys.stderr)
            return False

        # write config files
        with firebase_json_cfile.open('r', encoding='utf-8') as fr:
            json_text = fr.read()
            rendered_json = self.__synamic.render_string_template(
                json_text,
                # public_dir_name=output_cdir.basename <- was trouble maker as uploading is done by cd'ing into _outputs
            )
            with output_cdir.join(firebase_json_cfile.basename, is_file=True).open('w', encoding='utf-8') as fw:
                fw.write(rendered_json)

        with firebase_rc_cfile.open('r', encoding='utf-8') as fr:
            rc_text = fr.read()
            rendered_rc = self.__synamic.render_string_template(
                rc_text,
                project_name=firebase_projce_name
            )
            with output_cdir.join(firebase_rc_cfile.basename, is_file=True).open('w', encoding='utf-8') as fw:
                fw.write(rendered_rc)

        # upload
        cur_dt = datetime.datetime.utcnow().strftime('%y-%m-%d %H:%M:%S')
        command = ['firebase', 'deploy', '--only', 'hosting', '-m', f'Deploy @ {cur_dt}',
                   '--token', f'{firebase_ci_token}']
        cwd_bk = os.getcwd()
        os.chdir(output_cdir.abs_path)
        try:
            retcode = subprocess.call(command, shell=True)
        finally:
            os.chdir(cwd_bk)
            # TODO: remove firebase.json & .firebaserc from outputs dir after work is done? Yes, privacy/security issue.
        if retcode != 0:
            print(f'Upload was not successful, returned with code {retcode}', file=sys.stderr)
            return False
        else:
            return True



class Hashable:
    """A container to make hashable of any object so that
    we can use that object anywhere hashing is needed.
    e.g. set or dict operations"""
    __instance_count = 0

    def __init__(self, obj):
        __id = self.__instance_count = self.__instance_count + 1
        self.__id = __id
        self.__obj = obj

    @property
    def id(self):
        return self.__id

    @property
    def origin(self):
        return self.__obj

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__id == other.id

    def __hash__(self):
        return hash(self.__id)

    def __str__(self):
        return str(self.__obj)

    def __repr__(self):
        return repr(self.__obj)



import os
from synamic.core.services.event_system.events import EventSystem
from synamic.core.services.filesystem.path_tree import PathTree
from synamic.core.services.content.content_service import ContentService
from synamic.core.services.menu.menu_service import MenuService
from synamic.core.services.template.template_service import SynamicTemplateService
from synamic.core.services.image_resizer.image_resizer_service import ImageResizerService
from synamic.core.services.site_settings.site_settings import SiteSettingsService
from synamic.core.standalones.functions.decorators import loaded, not_loaded
from synamic.core.services.types.type_system import TypeSystem
from synamic.core.services.pre_processor import PreProcessorService
from synamic.core.services.tasks import TasksService
from synamic.core.services.marker import MarkerService
from synamic.core.services.user import UserService
from synamic.core.services.data import DataService
from synamic.core.default_data import DefaultDataManager
from synamic.core.contracts import SiteContract
from synamic.core.services.sitemap import SitemapService
from synamic.core.contracts import CDocType


def _install_default_services(site):
    self = site
    # Event Bus
    self.add_service('event_bus', EventSystem)

    # markers
    self.add_service('markers', MarkerService)
    # menus
    self.add_service('menus', MenuService)

    # user
    self.add_service('users', UserService)

    # templates service
    self.add_service('templates', SynamicTemplateService)
    # type system
    self.add_service('types', TypeSystem)

    # site settings
    self.add_service('site_settings', SiteSettingsService)

    # content service
    self.add_service('contents', ContentService)

    self.add_service('image_resizer', ImageResizerService)
    self.add_service('pre_processor', PreProcessorService)

    # data
    self.add_service('data', DataService)

    self.add_service('tasks', TasksService)

    # sitemap
    self.add_service('sitemap', SitemapService)


class _Site(SiteContract):
    def __init__(self, synamic, site_id, abs_path_to_site, parent_site=None, root_site=None):
        assert os.path.exists(abs_path_to_site)
        self.__synamic = synamic
        self.__site_id = site_id
        self.__abs_path_to_site = abs_path_to_site
        self.__parent_site = parent_site
        self.__root_site = root_site
        assert type(parent_site) in (self.__class__, type(None))
        assert type(root_site) in (self.__class__, type(None))

        # 1. >>> Check if the parent is the same type object.
        if self.__parent_site is not None:
            assert type(self.__parent_site) is type(self)
        # Root of all the synamic object must be None (Adam had no parent). Check it
        _root_parent = self.__parent_site
        while _root_parent:
            if _root_parent.parent is not None:
                _root_parent = _root_parent.parent
            else:
                _root_parent = None
        assert _root_parent is None
        # 1. <<<

        # Object Manager for site
        self.__object_manager_4_site = self.__synamic.object_manager.get_manager_for_site(self)

        # setting path tree
        self.__path_tree = PathTree.for_site(self)

        # Service container
        self.__services_container = {}
        _install_default_services(self)

        # Add path tree as service.
        # TODO: remove it later - keeping it now just to not crash the system.
        self.__services_container['path_tree'] = self.__path_tree

        self.__menus_wrapper = MenusWrapper(self.__object_manager_4_site)
        self.__users_wrapper = UsersWrapper(self.__object_manager_4_site)
        self.__data_wrapper = DataWrapper(self.__object_manager_4_site)

        self.__site_curl = None
        self.__is_loaded = False

    @property
    def synamic(self):
        return self.__synamic

    @property
    def id(self):
        return self.__site_id

    @property
    def abs_root_path(self):
        return self.__abs_path_to_site

    @property
    def parent(self):
        return self.__parent_site

    @property
    def has_parent(self):
        return False if self.__parent_site is None else True

    @property
    def root(self):
        return self.__root_site

    root_site = root

    @property
    def is_root(self):
        return self.__root_site is None

    def get_service(self, service_name, default=None, error_out=True):
        service = self.__services_container.get(service_name, None)
        if error_out:
            if service is None:
                raise Exception("Service with name %s could not be found." % service_name)
        if service is None:
            return default
        return service

    def add_service(self, service_name, service_class, init_args=(), init_kwargs=None):
        if init_kwargs is None:
            init_kwargs = {}
        assert service_name not in self.__services_container, 'Service with name `%s` already exists.' % service_name
        for existing_service_class in [type(service) for service in self.__services_container.values()]:
            if existing_service_class == service_class:
                raise Exception('Service instance already added. You are trying to add service with name'
                                '`%s` and type `%s`. Instance with name `%s` and the same type found.'
                                % (service_name, str(service_class), str(existing_service_class)))
        # Now do the real adding work.
        new_service = service_class(self, *init_args, **init_kwargs)
        self.__services_container[service_name] = new_service
        return new_service

    @not_loaded
    def load(self):
        for service in self.__services_container.values():
            service.load()
        self.__object_manager_4_site.load()
        self.__is_loaded = True
        return self

    @property
    def default_data(self) -> DefaultDataManager:
        return self.__synamic.default_data

    @property
    def system_settings(self):
        return self.__synamic.default_data.get_system_settings()

    @property
    def settings(self):
        return self.object_manager.get_site_settings()

    @property
    @loaded
    def curl(self):
        site_curl = self.__site_curl
        if site_curl is None:
            site_curl = self.__synamic.router.make_url(self, '', for_cdoctype=CDocType.HTML_DOCUMENT)
            # TODO: doctype can also be generated HTML! Solution?
            self.__site_curl = site_curl
        return site_curl

    @property
    def menu(self):
        return self.__menus_wrapper

    @property
    def user(self):
        return self.__users_wrapper

    @property
    def data(self):
        return self.__data_wrapper

    @property
    def path_tree(self):
        return self.__path_tree

    @property
    def event_bus(self) -> EventSystem:
        return self.__services_container['event_bus']

    @property
    def object_manager(self):
        return self.__object_manager_4_site

    @property
    def is_loaded(self) -> bool:
        return self.__is_loaded

    @property
    def site_wrapper(self):
        return _SiteWrapper(self)

    def __getattr__(self, service_name):
        return self.get_service(service_name, error_out=True)

    def __getitem__(self, service_name):
        return self.get_service(service_name, error_out=True)

    def __str__(self):
        return 'Site: ' + str(self.id) + ' <- ' + (str(self.parent.id.components) if self.has_parent else str(None)) \
               + ' [parent]'

    def __repr__(self):
        return repr(self.__str__())

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return other.id == other.id


class _SiteWrapper:
    """Wrapper for passing to templates and other places."""
    def __init__(self, site):
        self.__site = site
        self.__blacklisted_attrs = {'load', 'add_service', 'get_service', 'synamic', 'parent', 'root',
                                    'default_data', 'object_manager', 'parent', 'sites', 'router', 'builder',
                                    'site_wrapper'}

    def __getattr__(self, key):
        if key in self.__blacklisted_attrs:
            raise AttributeError('%s key is black listed' % key)
        else:
            try:
                return getattr(self.__site, key)
            except AttributeError:
                raise AttributeError('%s key was not found on site' % key)


class _SiteStuffWrapper:
    def __init__(self, origin_object):
        self.__origin_object__ = origin_object

    def get(self, key, default=None):
        return self.__origin_object__.get(key, default=default)

    def __getattr__(self, key):
        return self.get(key)

    def __getitem__(self, key):
        return self.get(key)


class MenusWrapper:
    def __init__(self, object_manager):
        self.__object_manager = object_manager

    def all(self):
        return self.__object_manager.get_menus()

    def get(self, key, default=None):
        return self.__object_manager.get_menu(key, default=default)

    def __getattr__(self, key):
        return self.get(key)

    def __getitem__(self, key):
        return self.get(key)


class UsersWrapper:
    def __init__(self, object_manager):
        self.__object_manager = object_manager

    def all(self):
        return self.__object_manager.get_users()

    def get(self, key, default=None):
        return self.__object_manager.get_user(key, default=default)

    def __getattr__(self, key):
        return self.get(key)

    def __getitem__(self, key):
        return self.get(key)


class DataWrapper:
    def __init__(self, object_manager):
        self.__object_manager = object_manager

    def get(self, key, default=None):
        return self.__object_manager.get_data(key, default=default)

    def __getattr__(self, key):
        return self.get(key)

    def __getitem__(self, key):
        return self.get(key)


from synamic.core.contracts import CDocType
from aiohttp import web
from synamic.exceptions import SynamicError


async def synamic_handler(request):
    synamic = request.app.synamic
    path_qs = f"{request.path}{'?' + request.query_string if request.query_string else ''}"
    print(f'Requesting: {path_qs}')
    # content = synamic.object_manager.router.get(path_qs)
    try:
        content = synamic.router.get_content(path_qs)

        if content is None:
            text = f"Not found for {path_qs}"
            mimetype = 'text/html'
            response = web.Response(text=text, content_type=mimetype)
        else:
            mimetype = content.mimetype
            print(f'Found content for {path_qs} as mimetype {mimetype} and cdoctype {content.cdoctype}')
            if CDocType.is_binary(content.cdoctype, not_generated=True):
                # serve static
                response = web.FileResponse(content.cpath.abs_path)
                response.content_type = mimetype

            elif CDocType.is_binary(content.cdoctype):
                # serve as stream.
                data = content.body_as_bytes
                response = web.Response(body=data, content_type=mimetype)
            else:
                # text content.
                with content.get_stream() as s:
                    data = s.read()
                text = data.decode('utf-8')
                response = web.Response(text=text, content_type=mimetype)
        return response

    except SynamicError as e:
        text = f"Synamic Error occured when requesting url: {path_qs}\n"\
               f"Error Details:\n{e.message}"
        mimetype = 'text/plain'
        response = web.Response(text=text, content_type=mimetype)
        return response



class UploadManager:
    def __init__(self, synamic):
        self.__synamic = synamic

        self.__uploaders_map = {}
        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    def load(self):
        self.__install_builtin_uploaders()
        self.__is_loaded = True

    def __install_builtin_uploaders(self):
        from .builtin_uploaders.firebase_uploader import FireBaseUploader
        self.add_uploader('firebase', FireBaseUploader)

    def add_uploader(self, name, uploader_class):
        assert type(uploader_class) is type
        assert name not in self.__uploaders_map
        self.__uploaders_map[name] = uploader_class(self.__synamic)

    def get_uploader(self, name, default=None):
        return self.__uploaders_map.get(name, default)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import abc


class ContentUrlContract(abc.ABC):
    """
        For YAML provided urls
            curl can be of string or another map.
            eg:
                curl: /my/path/to/file
            or:
                curl:
                    curl: /my/path/to/file
                    name: my_awesome_url_name  # this name must conform to identifier as in variables

            - a '/' will be added at the end of the urls if it is not there.
    """
    @property
    @abc.abstractmethod
    def absolute_url(self):
        """
            With the scheme and domain and path 
        """
        pass

    @property
    @abc.abstractmethod
    def path(self):
        """
        Root relative path
        Normalized path"""
        pass

    @property
    @abc.abstractmethod
    def generalized_path(self):
        """Generalized/lower path"""
        pass

    @property
    @abc.abstractmethod
    def generalized_real_path(self):
        """Generalized/lower real path"""
        pass

    @property
    @abc.abstractmethod
    def url_encoded_path(self):
        """
        :return: curl-encoded version of path
         
         Note: to_file_system_path does not need this because they do need to be present in html.
        """

    @property
    @abc.abstractmethod
    def real_path(self):
        """
        Must all, including this path, be normalized.
        
        e.g:
            path: me/one
            to_file_system_path: me/one/index.html
        
        e.g2:
            path: me/one/hi.html
            to_file_system_path: me/one/hi.html  # <the same>
        """

    @property
    @abc.abstractmethod
    def dir_components(self):
        """
            split to_file_system_path, not path, by forward slash '/'
            BUT this depends on is_file.
            If it is_file then the basename of it will be excluded
            
            Purpose:
                For creating directory depending on the array returned by this.
            
            Returns:
                A list or another ordered iterable.
        """

    @property
    @abc.abstractmethod
    def is_file(self):
        """
            is the resource, generated or static, indicates a file or directory
        """

    @property
    @abc.abstractmethod
    def is_dir(self):
        pass

    # @property
    # @abc.abstractmethod
    # def content(self):
    #     """
    #         Urls are associated only with some kind of content.
    #          This will return the content object (maybe, so that we can call render() on it later)
    #     """

    @property
    @abc.abstractmethod
    def to_file_path(self):
        """
            Must return a file name with os.sep specific to the running platform 
        """

    @abc.abstractmethod
    def create_auxiliary_url(self):
        pass



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

import re
import datetime
from collections import OrderedDict
import mimetypes
from synamic.core.contracts.content import CDocType
from synamic.core.standalones.functions.decorators import not_loaded
from synamic.core.services.content.static_content import StaticContent
from synamic.core.services.content.generated_content import GeneratedContent
from synamic.core.services.content.marked_content import MarkedContent
from synamic.core.services.content.paginated_content import PaginationPage, PaginatedContent
from synamic.core.contracts.cfields import CFieldsContract
from synamic.exceptions import SynamicSettingsError
from synamic.core.parsing_systems.marked_chapter import MarkedChapterParser


_invalid_url = re.compile(r'^[a-zA-Z0-9]://', re.IGNORECASE)


class ContentService:
    __slots__ = ('__site', '__is_loaded', '__contents_by_id', '__dynamic_contents', '__auxiliary_contents')

    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    @classmethod
    def is_type_cfields(cls, other):
        return isinstance(other, _ContentFields)

    @classmethod
    def is_type_synthetic_cfields(cls, other):
        return isinstance(other, _SyntheticContentFields)

    @classmethod
    def is_type_generated_content(cls, other):
        return isinstance(other, GeneratedContent)

    @classmethod
    def is_type_paginated_content(cls, other):
        return isinstance(other, PaginatedContent)

    @classmethod
    def is_type_pagination_page(cls, other):
        return isinstance(other, PaginationPage)

    def build_cfields(self, fields_syd, file_cpath):
        # get dir meta syd
        # """It should not live here as it is compile time dependency"""
        # each field from meta syd will be converted with individual content model and site type system.
        dir_meta_file_name = self.__site.system_settings['configs.dir_meta_file_name']
        _syd = self.__site.object_manager.empty_syd()
        parent_cpaths = file_cpath.parent_cpaths
        for dir_cpath in parent_cpaths:
            dir_meta_cfile = dir_cpath.join(dir_meta_file_name, is_file=True)
            if dir_meta_cfile.exists():
                dir_meta_syd = self.__site.object_manager.get_syd(dir_meta_cfile)
                _syd = _syd.new(dir_meta_syd)
        else:  # for else when loop ended normally without using break.
            fields_syd = _syd.new(fields_syd)

        # TODO: what is the document type???
        cdoctype = CDocType.HTML_DOCUMENT
        model_name = 'content'  # model name for contents is 'content' - a builtin model exists with this name.
                                # User can only override no existing field creating another model with the same name
                                # under site's meta model directory.
        model = self.__site.object_manager.get_model(model_name)
        curl = self.__site.object_manager.make_url_for_marked_content(
            file_cpath, path=fields_syd.get('path', None), slug=fields_syd.get('slug', None), for_cdoctype=cdoctype
        )
        mimetype = 'text/html'
        cfields = self.make_cfields(file_cpath, curl, model, cdoctype, mimetype, fields_syd)

        # TODO: can the following be more systematic and agile.
        # create markers that exists in cfields/contents but not in the system
        for marker_id in ('tags', 'categories'):
            if marker_id in fields_syd:
                marker = self.__site.object_manager.get_marker(marker_id)
                marks = self.__site.get_service('types').get_converter('marker#' + marker_id)(fields_syd[marker_id])
                for mark in marks:
                    if marker.get_mark_by_id(mark.id, None) is None:
                        marker.add_mark(mark)

        return cfields

    # def build_synthetic_cfields(self, curl, cdoctype, fields_map):
    #     return _SyntheticContentFields(self.__site, curl, cdoctype, fields_map)

    def build_md_content(self, file_cpath, cached_cfields):
        _, body_text = self.__site.object_manager.get_content_parts(file_cpath)
        # mime type guess
        content = MarkedContent(self.__site,
                                body_text,
                                cached_cfields)
        return content

    def build_paginated_md_content(self):
        # TODO: code it
        pass

    def build_static_content(self, path):
        path_tree = self.__site.path_tree
        file_cpath = path_tree.create_file_cpath(path)
        mimetype, _ = mimetypes.guess_type(file_cpath.basename)
        if mimetype is None:
            mimetype = 'octet/stream'  # TODO: improve guessing here.
        cdoctype = CDocType.BINARY_DOCUMENT
        curl = self.__site.object_manager.static_content_cpath_to_url(file_cpath, cdoctype)
        cfields_map = {}
        cfields = self.make_synthetic_cfields(curl, cdoctype, mimetype, cpath=file_cpath, fields_map=cfields_map)
        return StaticContent(
            self.__site,
            cfields
        )

    def build_generated_content(
            self,
            synthetic_cfields,
            file_content,
            source_cpath=None,
            render_callable=None):
        return GeneratedContent(self.__site, synthetic_cfields, file_content, source_cpath=source_cpath, render_callable=render_callable)

    def make_cfields(self, content_file_path, curl, cmodel, cdoctype, mimetype, raw_fileds):
        """Just makes an instance"""
        return _ContentFields(self.__site, content_file_path, curl, cmodel, cdoctype, mimetype, raw_fileds)

    def make_synthetic_cfields(self, curl, cdoctype, mimetype, cpath=None, fields_map=None):
        return _SyntheticContentFields(self.__site, curl, cdoctype, mimetype, cpath=cpath, fields_map=fields_map)

    def build_book_toc(self, toc_fn, book_cpath):
        toc_cpath = book_cpath.parent_cpath.join(toc_fn, is_file=True)
        book_toc = MarkedChapterParser(self.__site, book_cpath, toc_cpath).parse()
        return book_toc


class _ContentFields(CFieldsContract):
    def __init__(self, site, content_file_cpath, curl, cmodel, cdoctype, mimetype, raw_cfields):
        self.__site = site
        self.__content_file_cpath = content_file_cpath
        self.__curl = curl
        self.__cmodel = cmodel
        self.__cdoctype = cdoctype
        self.__mimetype = mimetype
        self.__raw_cfields = raw_cfields
        self.__converted_values = OrderedDict()

    def as_generated(self, curl, cdoctype=CDocType.GENERATED_HTML_DOCUMENT):
        """With cfields will use .set() and thus it will only affect converted values."""
        return _GeneratedContentFields(self.__site, curl, self.__cmodel, cdoctype, self)

    def __convert_pagination(self, pagination_field):
        object_manager = self.__site.object_manager
        site_settings = self.__site.settings
        per_page = per_page_4m_settings = site_settings['pagination_per_page']

        if isinstance(pagination_field, str):
            query_str = pagination_field
        else:
            query_str = pagination_field['query']
            per_page = pagination_field.get('per_page', None)
            if per_page is not None:
                if not isinstance(per_page, int):
                    raise SynamicSettingsError(f'Specified per page is not a positive integer number;'
                                               f' Type {type(per_page)} of per_page value {per_page}')
                per_page = per_page
            else:
                per_page = per_page_4m_settings
        fields = object_manager.query_cfields(query_str)
        origin_content = object_manager.get_marked_content(self.cpath)
        assert self is origin_content.cfields
        paginations, paginated_contents = PaginationPage.paginate_cfields(
            self.__site,
            origin_content,
            fields,
            per_page
        )
        return paginations, paginated_contents

    # def __convert_chapters(self, param):
    #     content_service = self.__site.get_service('contents')
    #     chapters = content_service.build_chapters(param)
    #     return chapters

    def get(self, key, default=None):
        raw_value = self.__raw_cfields.get(key, None)
        # generate value.
        if raw_value is None:
            if key == 'created_on':
                value = datetime.datetime.fromtimestamp(self.cpath.getctime())
            elif key == 'updated_on':
                updated_on = datetime.datetime.fromtimestamp(self.cpath.getmtime())
                created_on = self.get('created_on')
                assert created_on
                if updated_on < created_on:
                    updated_on = created_on
                value = updated_on
            elif key == 'chapter':
                # TODO: chapter is a generated prop. Validate that chapter is not in raw fields.
                value = None
                for book_toc in self.__site.object_manager.get_book_tocs():
                    book_chapter = book_toc.get_chapter(self.__content_file_cpath, default=None)
                    if book_chapter is not None:
                        value = book_chapter
                        break
            else:
                return default
        else:
            value = self.__converted_values.get(key, None)
            # convert value
            if value is None:
                # special conversions
                if key in ('pagination', 'book_toc', 'meta_tags', 'meta_description'):
                    if key == 'pagination':
                        pagination_field = raw_value
                        paginations, paginated_contents = self.__convert_pagination(pagination_field)
                        value = paginations[0]
                    elif key == 'book_toc':
                        content_service = self.__site.get_service('contents')
                        toc_fn = raw_value
                        book_toc = content_service.build_book_toc(toc_fn, self.__content_file_cpath)
                        value = book_toc
                    elif key in ('meta_tags', 'meta_description'):
                        lines = (line.strip() for line in raw_value.splitlines())
                        value = ' '.join(lines)
                    else:
                        raise Exception('Something is terribly wrong.')
                # normal conversions through converter or else raw value
                else:
                    model_field = self.__cmodel.get(key, None)
                    if model_field is not None:
                        value = model_field.converter(raw_value)
                    else:
                        value = raw_value
        self.__converted_values[key] = value
        return value

    def set(self, key, value):
        """Converted cfields will be affected only.
        Raw cfields will stay intact."""
        self.__converted_values[key] = value

    @property
    def raw(self):
        return self.__raw_cfields

    @property
    def keys(self):
        return tuple(set(self.__converted_values.keys()).union(set(self.__raw_cfields.keys())))

    @property
    def cdoctype(self):
        return self.__cdoctype

    @property
    def cmodel(self):
        return self.__cmodel

    @property
    def curl(self):
        return self.__curl

    @property
    def cpath(self):
        """Content file path"""
        return self.__content_file_cpath

    @property
    def mimetype(self):
        return self.__mimetype

    def __getitem__(self, key):
        return self.get(key, None)

    def __getattr__(self, key):
        return self.get(key, None)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__curl == other.curl

    def __hash__(self):
        return hash(self.__curl)


class _GeneratedContentFields(CFieldsContract):
    def __init__(self, site, curl, cmodel, cdoctype, origin_cfields):
        self.__site = site
        self.__curl = curl
        self.__cmodel = cmodel
        self.__cdoctype = cdoctype
        self.__origin_cfields = origin_cfields

        self.__overridden_cfields = {}

    def get(self, key, default=None):
        value = self.__overridden_cfields.get(key, None)
        if value is None:
            value = self.__origin_cfields.get(key, default)
        return value

    def set(self, key, value):
        """Converted cfields will be affected only.
        Raw cfields will stay intact."""
        self.__overridden_cfields[key] = value

    @property
    def raw(self):
        # TODO: what happens when generated field is changed but raw of origin cfields does not.
        return self.__origin_cfields.raw

    @property
    def keys(self):
        return tuple(set(self.__origin_cfields.keys()).union(set(self.__overridden_cfields.keys())))

    @property
    def cdoctype(self):
        return self.__cdoctype

    @property
    def cmodel(self):
        return self.__cmodel

    @property
    def curl(self):
        return self.__curl

    @property
    def cpath(self):
        raise NotImplemented

    @property
    def mimetype(self):
        return self.__origin_cfields.mimetype

    def __getitem__(self, key):
        return self.get(key, None)

    def __getattr__(self, key):
        return self.get(key, None)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__curl == other.curl

    def __hash__(self):
        return hash(self.__curl)


class _SyntheticContentFields(CFieldsContract):
    def __init__(self, site, curl, cdoctype, mimetype, cpath=None, fields_map=None):
        self.__site = site
        self.__curl = curl
        self.__cdoctype = cdoctype
        self.__mimetype = mimetype
        self.__cpath = cpath
        self.__fields_map = {} if fields_map is None else fields_map

    def get(self, key, default=None):
        return self.__fields_map.get(key, default)

    def set(self, key, value):
        self.__fields_map[key] = value

    @property
    def raw(self):
        return self.__fields_map

    @property
    def keys(self):
        return tuple(self.__fields_map.keys())

    @property
    def cdoctype(self):
        return self.__cdoctype

    @property
    def cmodel(self):
        raise NotImplemented

    @property
    def curl(self):
        return self.__curl

    @property
    def cpath(self):
        return self.__cpath

    @property
    def mimetype(self):
        return self.__mimetype

    def __getitem__(self, key):
        return self.get(key, None)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getattr__(self, key):
        return self.get(key, None)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__curl == other.curl

    def __hash__(self):
        return hash(self.__curl)



import enum

"""
* No event for path module loading.
* No event for instantiation
*   for example you are instantiating content/template/etc. module/service/etc., you cannot get/trigger event for this
*   because instantiation is the best place to register events without worrying whether the event being registered
*   is related to another module/service/etc. that need to be loaded before or after.
"""


@enum.unique
class EventTypes(enum.Enum):
    """
    These are all system events - not custom events.
    """

    SETTINGS_PRE_LOAD = 1
    SETTINGS_POST_LOAD = 2

    STATIC_CONTEND_ADDED = 3
    DYNAMIC_CONTENT_ADDED = 4

    # pre service/module load
    TEMPLATE_PRE_LOAD = "TEMPLATE_PRE"
    STATIC_PRE_LOAD = "STATIC_PRE"
    MODEL_PRE_LOAD = "MODEL_PRE"
    CONTENT_PRE_LOAD = "CONTENT_SERVICE_PRE"
    MENU_PRE_LOAD = "MENU_PRE"
    TAGS_PRE_LOAD = "TAGS_PRE"
    CATEGORIES_PRE_LOAD = "CATEGORIES_PRE"
    NULL_PRE_LOAD = "NULL_PRE"

    # post service/module load
    TEMPLATE_POST_LOAD = "TEMPLATE_POST"
    STATIC_POST_LOAD = "STATIC_POST"
    MODEL_POST_LOAD = "MODEL_POST"
    CONTENT_POST_LOAD = "CONTENT_SERVICE_POST"
    MENU_POST_LOAD = "MENU_POST"
    TAGS_POST_LOAD = "TAGS_POST"
    CATEGORIES_POST_LOAD = "CATEGORIES_POST"
    NULL_POST_LOAD = "NULL_POST"

    # synamic
    SYNAMIC_PRE_INIT = "SYNAMIC_PRE_INIT"
    SYNAMIC_POST_INIT = "SYNAMIC_POST_INIT"
    SYNAMIC_PRE_LOAD = "SYNAMIC_PRE_LOAD"
    SYNAMIC_POST_LOAD = "SYNAMIC_POST_LOAD"

    # STREAM : I guess I will not implement this event.
    STREAM_BEFORE = "STREAM_BEFORE"
    STREAM_AFTER = "STREAM_AFTER"

    # BUILD
    PRE_BUILD = "BUILD_PRE"
    POST_BUILD = "BUILD_POST"

    # SERVE
    PRE_SERVE = "PRE_SERVE"
    POST_SERVE = "POST_SERVE"

    # DEPLOY
    PRE_DEPLOY = "PRE_DEPLOY"
    POST_DEPLOY = "POST_DEPLOY"

    # System
    SYSTEM_PRE_START = "SYSTEM_PRE_START"
    #  no post start



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import re
import datetime
from synamic.exceptions import SynamicInvalidDateTimeFormat


_datetime_pattern = re.compile(r"^\s*(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<date>\d{1,2})\s*(\s+"
                               r"(?P<hour>\d{1,2}):(?P<minute>\d{1,2})(:(?P<second>\d{1,2}))?\s*"
                               r"(?P<am_pm>AM|PM)?)?\s*$"
                               , re.I)  # format: YEAR-MONTH-DATE HOUR:MIN:SECOND AM/PM

_date_pattern = re.compile(r"^\s*(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<date>\d{1,2})\s*$", re.I)  # format: YEAR-MONTH-DATE

_time_pattern = re.compile(r"\s*(?P<hour>\d{1,2}):(?P<minute>\d{1,2})(:(?P<second>\d{1,2}))?\s*(?P<am_pm>AM|PM)?\s*$", re.I)  # HOUR:MIN:SECOND AM/PM

ReMatchType = type(re.match('', ''))


class DtPatterns:
    datetime = _datetime_pattern
    date = _date_pattern
    time = _time_pattern


def _s2int(txt):
    txt = txt.strip()
    txt = txt.lstrip('0')
    if txt == '':
        return 0
    return int(txt)


def parse_datetime(txt) -> datetime.datetime:
    if isinstance(txt, datetime.datetime):
        return txt
    elif isinstance(txt, str):
        txt = txt.strip()
        m = _datetime_pattern.match(txt)
        if not m:
            raise SynamicInvalidDateTimeFormat(f"DateTime format did not match for text: {txt}")
    elif type(txt) is ReMatchType:
        m = txt
        assert m.re is DtPatterns.datetime
    else:
        raise SynamicInvalidDateTimeFormat(f"Invalid DateTime data for conversion provided, provided type is "
                                           f"{type(txt)} and string representation is {str(txt)}")

    year = _s2int(m.group('year'))
    month = _s2int(m.group('month'))
    date = _s2int(m.group('date'))
    hour = _s2int(m.group('hour')) if m.group('hour') else 0
    minute = _s2int(m.group('minute')) if m.group('minute') else 0
    second = _s2int(m.group('second')) if m.group('second') else 0
    am_pm = m.group('am_pm').upper() if isinstance(m.group('am_pm'), str) else m.group('am_pm')

    # 24 hour conversion
    if am_pm:
        if not hour < 13:
            raise SynamicInvalidDateTimeFormat(f"Hour cannot be greater than 12 when am/pm specified: {txt}")
        if am_pm.upper() == "PM":
            hour = 12 + hour

    dt = datetime.datetime(year, month, date, hour, minute, second)
    return dt


def parse_date(txt) -> datetime.date:
    if isinstance(txt, datetime.date):
        return txt
    elif isinstance(txt, str):
        txt = txt.strip()
        m = _date_pattern.match(txt)
        if not m:
            raise SynamicInvalidDateTimeFormat(f"Date format did not match for text: {txt}")
    elif type(txt) is ReMatchType:
        m = txt
        assert m.re is DtPatterns.date
    else:
        raise SynamicInvalidDateTimeFormat(f"Invalid Date data for conversion provided, provided type is {type(txt)}"
                                           f" and string representation is {str(txt)}")
    year = _s2int(m.group('year'))
    month = _s2int(m.group('month'))
    date = _s2int(m.group('date'))

    dt = datetime.date(year, month, date)
    return dt


def parse_time(txt) -> datetime.time:
    if isinstance(txt, datetime.time):
        return txt
    elif isinstance(txt, str):
        txt = txt.strip()
        m = _time_pattern.match(txt)
        if not m:
            raise SynamicInvalidDateTimeFormat(f"Time format did not match with text: {txt}")
    elif type(txt) is ReMatchType:
        m = txt
        assert m.re is DtPatterns.time
    else:
        raise SynamicInvalidDateTimeFormat(f"Invalid DateTime data for conversion provided, provided type is "
                                           f"{type(txt)} and string representation is {str(txt)}")
    hour = _s2int(m.group('hour')) if m.group('hour') else 0
    minute = _s2int(m.group('minute')) if m.group('minute') else 0
    second = _s2int(m.group('second')) if m.group('second') else 0
    am_pm = m.group('am_pm').upper() if isinstance(m.group('am_pm'), str) else m.group('am_pm')

    # 24 hour conversion
    if am_pm:
        if am_pm.upper() == "PM":
            hour = 12 + hour

    t = datetime.time(hour, minute, second)
    return t



import os
import os.path
from synamic.core.contracts import BaseFsBackendContract
from synamic.exceptions import SynamicFSError


class FileSystemBackend(BaseFsBackendContract):
    def open(self, path, *args, **kwargs):
        try:
            fo = open(path, *args, **kwargs)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during opening path {path}):\n'
                f'{str(e)}'
            )
        return fo

    def exists(self, path):
        try:
            res = os.path.exists(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during exists() on path {path}):\n'
                f'{str(e)}'
            )
        return res

    def is_file(self, path):
        try:
            res = os.path.isfile(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during checking if the path is a file path {path}):\n'
                f'{str(e)}'
            )
        return res

    def is_dir(self, path):
        try:
            res = os.path.isdir(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during checking if the path is a directory path {path}):\n'
                f'{str(e)}'
            )
        return res

    def listdir(self, path):
        try:
            res = os.listdir(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during listing path: {path}):\n'
                f'{str(e)}'
            )
        return res

    def makedirs(self, path):
        try:
            res = os.makedirs(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during making directory with path: {path}):\n'
                f'{str(e)}'
            )
        return res

    def getmtime(self, path):
        try:
            res = os.path.getmtime(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during getmtime on path: {path}):\n'
                f'{str(e)}'
            )
        return res

    def getctime(self, path):
        try:
            res = os.path.getctime(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during getctime on path: {path}):\n'
                f'{str(e)}'
            )
        return res

    def remove(self, path):
        try:
            res = os.remove(path)
        except (OSError, IOError) as e:
            raise SynamicFSError(
                f'Synamic File System Error (occurred during remove on path: {path}):\n'
                f'{str(e)}'
            )
        return res



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

from jinja2 import nodes, lexer
from jinja2.ext import Extension


class GetCExtension(Extension):
    tags = {'geturl', 'getc'}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(
            site_object=None
        )

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        token = parser.stream.expect(lexer.TOKEN_STRING)
        url_name_or_id = nodes.Const(token.value)
        call = self.call_method('_get_url', [url_name_or_id], lineno=lineno)
        return nodes.Output([call], lineno=lineno)

    def _get_url(self, parameter):
        url = self.environment.site_object.object_manager.getc(parameter)
        return url


class ResizeImageExtension(Extension):
    tags = {'resize'}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(
            site_object=None
        )

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        path, width, height = args
        call = self.call_method('_resize', [path, width, height], lineno=lineno)
        return nodes.Output([call], lineno=lineno)

    def _resize(self, path, width, height):
        cnt = self.environment.site_object.object_manager.resize_image(path, width, height)
        return cnt.curl.path_as_str



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import http.server
import re
import sys
from urllib.parse import unquote

from synamic.core.synamic.router.url import ContentUrl
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
import socket


class SynamicDevServerRequestHandler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        cfg = self.server.synamic_config
        paths = self.path.split("?")
        if len(paths) > 1:
            path = "".join(paths[:-1])
            print("self.path %s and path %s" % (self.path, path))
        else:
            path = self.path

        if re.match(r'/?--sys--/?$', path):
            self.send_response(200, message="Ok")
            self.send_header('Content-Type', 'text/html')
            self.end_headers()
            self.wfile.write(b"""
            <body>
                <br/><br/><br/>
                <a href="/--sys--/reload/"> Reload.. </a>
                <br/><br/><br/>
            </body>
            """)
            return

        elif re.match(r'/?--sys--/reload/?$', path):
            referer = goto = self.headers['referer']
            if referer:
                if re.match(r'/?--sys--/reload/?$', referer):
                    goto = '/--sys--/'
            else:
                goto = '/--sys--/'

            print("Reloading...")
            self.signal_reload()  # sync reload
            print("Reloaded!")
            self.send_response(302, message="Redirect")
            self.send_header('Location', goto)
            self.end_headers()
            return

        path = unquote(path)

        curl = ContentUrl(cfg, path)

        cont = self.server.synamic_config.get_content_by_content_url(curl)
        if cont is None:
            # content urls are coditionally slashed, let's try other way
            slug = path
            if not re.match(r'.*\.[a-z0-9]+$', slug, re.I):
                if not slug.endswith('/'):
                    slug += '/'
            curl = ContentUrl(cfg, slug)
            cont = self.server.synamic_config.get_content_by_content_url(curl)
        if cont is not None:
            self.send_response(200, message="OK")
            self.end_headers()
            stream = cont.get_stream()
            while True:
                byts = stream.read(60*1024)
                if not byts:
                    break
                else:
                    self.wfile.write(byts)
            stream.close()
        else:
            self.send_response(404, message="Not FounD")
            self.end_headers()
            self.wfile.write(('Not found %s' % self.path).encode('utf-8'))

    def log_request(self, code='-', size='-'):
        print("LOG REQUEST: %s - %s - %s" % (self.path, code, size))

    def signal_stop(self):
        self.server.signaled_stop = True

    def signal_reload(self):
        self.server.signaled_reload = True


class SynamicServer(http.server.HTTPServer):
    def __init__(self, config, *args, **kwargs):
        self.synamic_config = config
        self.signaled_reload = False
        self.signaled_stop = False
        super().__init__(*args, **kwargs)


def serve(synamic, port):
    server = SynamicServer(synamic, ('localhost', port), SynamicDevServerRequestHandler)
    _addr = server.server_address[0] + ":" + str(server.server_address[1])
    print("SERVE(): Server starting on %s" % _addr)

    observer = Observer()
    observer.schedule(MyHandler(server), path=synamic.site_root, recursive=True)
    observer.start()

    while server.signaled_stop is False and server.signaled_reload is False:
        try:
            server.handle_request()
        except:
            print("Shutdown exception?")
            break

    print("Stopping observer")
    observer.stop()
    print("joining...")
    observer.join()
    print('join exit...')

    # sys.stderr.flush()
    # sys.stdout.flush()

    if server.signaled_stop:
        return False
    elif server.signaled_reload:
        return True
    else:
        return True


class MyHandler(FileSystemEventHandler):
    def __init__(self, server, *args, **kwargs):
        self.__server = server
        self.__shutdown_in_progress = False
        super(*args, **kwargs)

    def process(self, event):
        """
        event.event_type 
            'modified' | 'created' | 'moved' | 'deleted'
        event.is_directory
            True | False
        event.src_path
            path/to/observed/file
        """
        if self.__shutdown_in_progress:
            print("Don't disturb, shutdown in progress")
            print("Event: %s for: %s" % (event.event_type, event.src_path))
            raise Exception("It should not be here anymore")

        self.__shutdown_in_progress = True
        print("Event: %s for: %s" % (event.event_type, event.src_path))
        print("Shutting down server...")
        try:
            self.__server.socket.shutdown(socket.SHUT_RDWR)
        except:
            pass
        self.__server.socket.close()
        # self.__server.server_close()
        print("...was shutdown")
        self.__server.signal_reload = True
        sys.exit(0)

    def on_any_event(self, event):
        self.process(event)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import os
import jinja2
from synamic.core.contracts import SynamicContract
from synamic.core.synamic.sites.sites import Sites
from synamic.core.synamic.router import RouterService
from synamic.core.default_data import DefaultDataManager
from synamic.core.standalones.functions.decorators import loaded, not_loaded
from synamic.core.object_manager import ObjectManager
from synamic.core.services.filesystem.path_tree import PathTree
from synamic.exceptions import get_source_snippet_from_text, SynamicTemplateError
from synamic.core.upload_manager.upload_manager import UploadManager


class Synamic(SynamicContract):
    def __init__(self, root_site_root):
        assert os.path.exists(root_site_root)
        assert os.path.isdir(root_site_root)
        self.__root_site_root = root_site_root

        # Default Config Manager
        self.__default_data = DefaultDataManager()

        # Object Manager
        self.__object_manager = ObjectManager(self)

        self.__path_tree = PathTree(self)

        self.__sites = Sites(self, self.__root_site_root)
        self.__router = RouterService(self)

        self.__upload_manager = UploadManager(self)

        # env
        self.__env = {
            'backend': 'file'
        }

        # dev server param
        self.__dev_params = {}

        self.__is_loaded = False

    @property
    def env(self):
        return self.__env

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__sites.load()  # TODO: sites should be loaded individually.
        self.__upload_manager.load()
        self.__is_loaded = True

    @property
    def default_data(self):
        return self.__default_data

    @property
    def object_manager(self):
        return self.__object_manager

    @property
    def system_settings(self):
        return self.__default_data.get_system_settings()

    @property
    # @loaded
    def sites(self):
        return self.__sites

    @property
    def abs_root_path(self) -> str:
        return self.__root_site_root

    @property
    def path_tree(self):
        return self.__path_tree

    # @loaded
    @property
    def router(self):
        return self.__router

    @property
    def event_system(self):
        raise NotImplemented

    @property
    def upload_manager(self):
        return self.__upload_manager

    def set_dev_params(self, **paras):
        self.__dev_params.update(paras)

    @property
    def dev_params(self):
        return self.__dev_params.copy()

    def render_string_template(self, text, context=None, **ctx):
        return render_string_template(text, context=context, **ctx)


def render_string_template(text, context=None, **ctx):
    template = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(text)
    if context is None:
        context = {}
    context.update(ctx)
    try:
        rendered_text = template.render(context)
    except jinja2.TemplateError as e:
        # TODO: error reporting with source - it is not a file, it is a string.
        raise SynamicTemplateError(e)
    else:
        return rendered_text



"""
Limited markdown parser only to parse chapters, nothing more nothing less.
"""
import re
from synamic.exceptions import (
    SynamicError,
    SynamicErrors,
    SynamicPathDoesNotExistError,
    get_source_snippet_from_text
)

__all__ = ['MarkedChapterParser']


# errors
class MismatchedListType(SynamicError):
    pass


class InvalidIndentation(SynamicError):
    pass


class TabsNotAllowed(SynamicError):
    pass


class InvalidLevel(SynamicError):
    pass


class InvalidLine(SynamicError):
    pass


class _Patterns:
    # list item
    list_item = re.compile(r'^(?P<white_spaces>\s*)(?P<item_marker>[*+-]|[0-9]+\. )[ \t]*(?P<item_data>.+)')

    # link pattern
    link = re.compile(
        r"""
        ^[ ]*
        \[
            (?P<link_text>[^\[\]\(]+)
        \]
        \(  
            (
                (?P<link_src>[^()'"]+?)
                ([ \t]+   ('|")   (?P<link_title>.+)   \1)?
            )
        \)
        [ ]*$
        """, re.VERBOSE)

    # blanks
    blanks = re.compile(r'^[ \t]*$')

    # part
    part_header = re.compile(r'^#(?P<text>[^#].*)')


class MarkedChapterParser:
    def __init__(self, site, book_cpath, toc_cpath):
        with toc_cpath.open('r', encoding='utf-8') as fr:
            toc_text = fr.read()

        self.__text = toc_text
        self._chapter_toc = ChaptersTOC(site, book_cpath, toc_cpath)

        self.__item_list_stack = []
        self.__current_blank_lines = 0

    def on_list_item(self, idx, white_spaces, item_marker, item_data):
        # calculate white space
        level = len(white_spaces) // 4
        # print(f'Level: {level}: @.{idx} {item_data} -> {type(item_data)}')

        # determine type of list: when needed
        if item_marker in ('*', '-', '+'):
            # TODO: check marker be equal ... if not start another list closing the last one.
            # check indent too.
            is_ordered = False
        else:
            is_ordered = True

        # CALCULATE PARENT & stack ops #
        # if there were blank lines preceding
        if self.__current_blank_lines:
            self.__current_blank_lines = 0
            # go to the beginning no matter how deep it is nested.
            self.__pop_until_level(0)
            ch_list = ChapterList(self._chapter_toc, parent=None, ordered=is_ordered)
            self.__item_list_stack.append(ch_list)
            parent_list = ch_list
        # ok add the item/list to existing item/list. OR create new one and add
        else:
            # create the very first list & add item to it.
            if len(self.__item_list_stack) == 0:
                if level != 0:
                    raise InvalidLevel(
                        f'Level of the current item is invalid.\n' +
                        f'Source:\n' +
                        get_source_snippet_from_text(text, idx + 1)
                    )

                ch_list = ChapterList(self._chapter_toc, parent=None, ordered=is_ordered)
                self.__item_list_stack.append(ch_list)
                parent_list = ch_list
            # we will use existing list.
            else:
                # assert level > 0, f'Level: {level}'
                # go up one level (if needed) & find the parent matching level if current level is nested
                # one level deeper
                li_parent = self.__item_list_stack[-1]
                # find the real parent list and add item.
                if level == li_parent.level:
                    # self.__pop_until_level(level)
                    parent_list = self.__item_list_stack[-1]
                # this item is indented one level
                elif level - 1 == li_parent.level:
                    list_item = self.__item_list_stack[-1].get_last_item()
                    parent_list = ChapterList(
                        self._chapter_toc,
                        parent=list_item,
                        ordered=is_ordered
                    )
                    list_item.set_list(parent_list)
                    self.__item_list_stack.append(parent_list)
                # this item is dedented any level
                elif level < li_parent.level:
                    self.__pop_until_level(level)
                    parent_list = self.__item_list_stack[-1]

                else:
                    raise InvalidIndentation(
                        'Invalid indentation:: ' + f'Level: {level}, Parent Level: {li_parent.level}\n' +
                        f'Source:\n' +
                        get_source_snippet_from_text(text, idx + 1)
                    )

        # type of current item & li parent must be the same
        if parent_list.is_ordered != is_ordered:
            raise MismatchedListType(
                'Ordered & Unordered list item mixed. Previous'
                ' and current ordering does not match.\n' +
                f'Source:\n' +
                get_source_snippet_from_text(text, idx + 1)
            )

        # OK, now, set the item
        list_item = ListItem(self._chapter_toc, parent_list, item_data)
        if isinstance(item_data, ChLink):
            try:
                ch_link = item_data
                chapter = self._chapter_toc.add_chapter(ch_link)
                # set real url to the ch_link
                ch_link.__set_url__(chapter.url)
            except SynamicError as e:
                raise SynamicErrors(
                    'Error during adding url to chapter toc\n' +
                    f'Source:\n' +
                    get_source_snippet_from_text(self.__text, idx + 1)
                    ,
                    e
                )
        parent_list.add_child(list_item)

    def on_blank_line(self, idx):
        self.__pop_until_level(0)

        self.__current_blank_lines += 1
        # print(f'@.{idx} ... blank ...')

    def on_part(self, idx, part_text):
        self.__pop_until_level(0)

        last_part = self._chapter_toc.get_last_part()
        if last_part is None:
            empty_part_text = ''
            last_part = ChapterPart(empty_part_text)
            self._chapter_toc.add_part(
                last_part
            )
        for ch_list in self.__item_list_stack:
            last_part.add_chapter_list(ch_list)

        # current part
        current_part = ChapterPart(part_text)
        self._chapter_toc.add_part(current_part)

        self.__item_list_stack.clear()

        # print(f'@.{idx} part: {part_text}')

    def on_end(self):
        self.__pop_until_level(0)

        last_part = self._chapter_toc.get_last_part()
        if last_part is None:
            empty_part_text = ''
            last_part = ChapterPart(empty_part_text)
            self._chapter_toc.add_part(
                last_part
            )
            self._chapter_toc.__on_mutation_done__()
        for ch_list in self.__item_list_stack:
            last_part.add_chapter_list(ch_list)
        self.__item_list_stack.clear()

    def on_other(self, idx, text):
        raise InvalidLine(
            'Invalid data that matched no pattern.\n'
            f'Source:\n' +
            get_source_snippet_from_text(self.__text, idx + 1)
        )

    def __pop_until_level(self, level=0):
        while len(self.__item_list_stack) and \
                self.__item_list_stack[-1].level > level:
            self.__item_list_stack.pop()

    def parse(self):
        return self.__lex(self.__text, self)

    def __lex(self, text: str, parser):
        lines = text.splitlines()

        for idx, line in enumerate(lines):
            blank_line_match = _Patterns.blanks.match(line)
            list_item_match = _Patterns.list_item.match(line)
            part_header_match = _Patterns.part_header.match(line)

            # skip blank lines
            if blank_line_match:
                parser.on_blank_line(idx)
                continue
            if list_item_match:
                # when it is a list item
                white_spaces = list_item_match.group('white_spaces')
                item_marker = list_item_match.group('item_marker')
                item_data = list_item_match.group('item_data')

                # tab validation
                if '\t' in white_spaces:
                    raise TabsNotAllowed(
                        'Tabs not allowed for nesting list items.\n' +
                        'Source:\n' +
                        get_source_snippet_from_text(text, idx + 1)
                    )

                # indentation validation
                if len(white_spaces) % 4 != 0:
                    raise InvalidIndentation(
                        f'Indentation must be multiple of zero or more 4 space characters\n' +
                        'Source:\n' +
                        get_source_snippet_from_text(text, idx + 1)
                    )

                # when this list item contains link pattern.
                link_match = _Patterns.link.match(item_data)
                if link_match:
                    link_text = link_match.group('link_text')
                    link_src = link_match.group('link_src')
                    link_title = link_match.group('link_title')
                    item_data = ChLink(self._chapter_toc, link_text, link_src, link_title)
                parser.on_list_item(idx, white_spaces, item_marker, item_data)

            elif part_header_match:
                part_text = part_header_match.group('text')
                parser.on_part(idx, part_text)
            else:
                parser.on_other(idx, line)
        parser.on_end()
        toc = self._chapter_toc
        # self.__init__(toc.site, toc_cpath, self.__text)
        return toc


class ChapterList:
    def __init__(self, chapter_toc, parent=None, ordered=True):
        assert isinstance(chapter_toc, ChaptersTOC)
        assert isinstance(parent, (type(None), ListItem))
        assert parent is not self
        self._chapter_toc = chapter_toc
        self._parent = parent
        self._ordered = ordered
        self._items = []

    @property
    def chapter_toc(self):
        return self._chapter_toc

    @property
    def is_ordered(self):
        return self._ordered

    @property
    def is_unordered(self):
        return not self._ordered

    @property
    def has_parent(self):
        return self._parent is not None

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return tuple(self._items)

    @property
    def has_children(self):
        return len(self._items) > 0

    def add_child(self, item):
        assert isinstance(item, ListItem)
        assert item.parent is self
        self._items.append(item)

    def get_last_item(self):
        return self._items[-1]

    @property
    def is_list(self):
        return True

    @property
    def level(self):
        _max = 10
        level = 0
        parent = self.parent
        while parent is not None:
            parent = parent.parent
            if not isinstance(parent, (ListItem, type(None))):  # not checking None was the bug.
                level += 1
            assert level < _max, 'Ridiculous nesting or bug'
        return level

    def as_html(self, added_level=None):
        descendants = self.children
        res = []
        level = self.level if added_level is None else self.level + added_level
        # if isinstance(self.parent, ListItem):
        #     level += 1
        for child in descendants:
            res.append(child.as_html(added_level=1 + added_level if added_level is not None else 1))

        if self.is_ordered:
            return \
                (level * '    ') + '<ol class="book-toc-chapter-list">\n' + \
                '\n'.join(res) + '\n' +\
                (level * '    ') + '</ol>'
        else:
            return \
                (level * '    ') + '<ul class="book-toc-chapter-list">\n' + \
                '\n'.join(res) + '\n' +\
                (level * '    ') + '</ul>'

    def __str__(self):
        descendants = self.children
        res = []
        level = self.level
        for child in descendants:
            res.append(level * '    ' + str(child))
        return '\n'.join(res)

    def __repr__(self):
        return repr(self.__str__())


class ListItem:
    def __init__(self, chapter_toc, parent, data):
        assert isinstance(chapter_toc, ChaptersTOC)
        assert isinstance(parent, ChapterList)
        assert isinstance(data, (str, ChLink)), f'Type: {type(data)}'
        self._chapter_toc = chapter_toc
        self.__parent = parent
        self.__data = data

        self.__child_list = None

    @property
    def chapter_toc(self):
        return self._chapter_toc

    @property
    def level(self):
        return self.parent.level

    @property
    def data(self):
        return self.__data

    @property
    def has_parent(self):
        return self.__parent is not None

    @property
    def parent(self):
        return self.__parent

    @property
    def is_item(self):
        return True

    @property
    def has_list(self):
        return self.__child_list is not None

    @property
    def child_list(self):
        return self.__child_list

    def set_list(self, l):
        assert self.__child_list is None
        assert isinstance(l, ChapterList)
        assert l.parent is self
        self.__child_list = l

    def as_html(self, added_level=None):
        item_data = f'{self.__data}' if isinstance(self.__data, str) else self.__data.as_html()
        # item_data = '   ' * self.level + item_data
        level = self.level if added_level is None else self.level + added_level
        if not self.__child_list:
            return \
                ('    ' * level) + '<li>' + item_data + '</li>'
        else:
            return \
                '    ' * level + '<li>' + item_data + '\n' + \
                self.__child_list.as_html(added_level=added_level) + '\n' + \
                '    ' * level + '</li>'

    def __str__(self):
        res = []
        res.append(str(self.__data))
        if self.__child_list:
            res.append(str(self.__child_list))
        return '\n'.join(res)

    def __repr__(self):
        return repr(self.__str__())


class ChLink:
    def __init__(self, chapter_toc, link_text, link_src, link_title):
        assert isinstance(chapter_toc, ChaptersTOC)
        self._chapter_toc = chapter_toc
        self.__link_text = link_text
        self.__link_src = link_src
        self.__link_title = link_title
        self.__url = None

    def __set_url__(self, url):
        assert self.__url is None
        self.__url = url

    @property
    def link_text(self):
        return self.__link_text

    @property
    def link_src(self):
        return self.__url if self.__url is not None else self.__link_src

    @property
    def link_title(self):
        return self.__link_title

    def as_html(self):
        title_text = '' if not self.link_title else f' title="{self.link_title}"'
        return f'<a href="{self.link_src}"{title_text}>{self.link_text}</a>'

    def __str__(self):
        return f'{self.link_text}: {self.link_src}, {self.link_title}'

    def __repr__(self):
        return repr(self.__str__())


class Chapter:
    def __init__(self, book_toc, title, curl, anchor_title, previous=None):
        self.__book_toc = book_toc
        self.__title = title
        self.__curl = curl
        self.__anchor_title = '' if anchor_title is None else anchor_title
        self.__previous = previous
        self.__next = None

    @property
    def book(self):
        return self.__book_toc.book

    @property
    def book_toc(self):
        return self.__book_toc

    @property
    def curl(self):
        return self.__curl

    @property
    def url(self):
        return self.__curl.url

    @property
    def title(self):
        return self.__title

    @property
    def anchor_title(self):
        return self.__anchor_title

    @property
    def as_anchor(self):
        if self.anchor_title:
            return f'<a href="{self.url}" title="{self.anchor_title}">{self.title}</a>'
        else:
            return f'<a href="{self.url}">{self.title}</a>'

    @property
    def as_a(self):
        return self.as_anchor

    @property
    def next(self):
        """Next chapter object"""
        return self.__next

    @property
    def previous(self):
        """Previous chapter object"""
        return self.__previous

    @property
    def prev(self):
        return self.previous

    def __set_next__(self, next_chapter):
        assert isinstance(next_chapter, self.__class__)
        assert self.__next is None
        self.__next = next_chapter

    def __str__(self):
        return self.as_anchor

    def __repr__(self):
        return repr(self.__str__())


class ChapterPart:
    def __init__(self, name=''):
        """if name is none then """
        assert isinstance(name, str)
        self._name = name
        self._chapter_lists = []

    def add_chapter_list(self, ch_list):
        assert isinstance(ch_list, ChapterList)
        self._chapter_lists.append(ch_list)

    @property
    def chapter_lists(self):
        return tuple(self._chapter_lists)

    def as_html(self, added_level=None):
        added_level = 0 if added_level is None else added_level
        header = '    ' * added_level + ('' if not self._name else
                                         f"<div class='book-toc-part-header'>{self._name}</div>")
        return \
            f'\n{header}\n' +\
            '\n'.join(chl.as_html(added_level=added_level) for chl in self._chapter_lists) + '\n'

    def __str__(self):
        return '\nPart: ' + self._name + '>\n' + '\n'.join(str(chl) for chl in self._chapter_lists)

    def __repr__(self):
        return repr(self.__str__())


class ChaptersTOC:
    def __init__(self, site, book_cpath, toc_cpath):
        self._site = site
        self.__book_cpath = book_cpath
        self.__chapter_parts = []
        self.__cpath_to_chapter = {}

        self.__toc_cpath = toc_cpath
        self.__toc_parent_cpath = self.__toc_cpath.parent_cpath

        # tmp
        self.__chapter_cpath_tuples = []

        self.__is_done = False

    @property
    def site(self):
        return self._site

    @property
    def book_cpath(self):
        return self.__book_cpath

    @property
    def book_curl(self):
        return self._site.object_manager.get_cfields(self.__book_cpath).curl

    @property
    def book_url(self):
        return self._site.object_manager.get_cfields(self.__book_cpath).curl.url

    @property
    def book_cfields(self):
        return self._site.object_manager.get_cfields(self.__book_cpath)

    @property
    def book(self):
        return self._site.object_manager.get_marked_content(self.__book_cpath)

    @property
    def chapter_cpaths(self):
        return tuple(self.__cpath_to_chapter.keys())

    def add_part(self, part):
        assert not self.__is_done
        assert isinstance(part, ChapterPart)
        self.__chapter_parts.append(part)

    def add_chapter(self, ch_link: ChLink):
        assert not self.__is_done
        assert isinstance(ch_link, ChLink)
        href = ch_link.link_src
        if href.startswith('/'):
            raise SynamicError(f'Chapter link must be relative: {href}')
        marked_extensions = self.site.system_settings['configs.marked_extensions']
        if not href.endswith(marked_extensions):
            raise SynamicError(f'Chapter link must be a relative path to a file with an extension of marked content'
                               f' {marked_extensions}: {href}')
        if href.endswith('/'):
            raise SynamicError(f'Chapter link must be a relative file name to a marked content: {href}')

        content_cpath = self.__toc_parent_cpath.join(href, is_file=True)
        if not content_cpath.exists():
            raise SynamicPathDoesNotExistError(
                f'CPath referred to by chapter link {href} does not exist: {content_cpath.abs_path}'
            )
        if content_cpath in self.__cpath_to_chapter:
            raise SynamicError(
                f'CPath referred to by chapter link {href} already added to the toc: {content_cpath.abs_path}'
            )

        # process chapter cpath
        prev_chapter = None if not self.__chapter_cpath_tuples else self.__chapter_cpath_tuples[-1][0]
        cfields = self._site.object_manager.get_cfields(content_cpath)
        if cfields is None:
            raise SynamicError(
                f'CFields not found for content path {content_cpath.relative_path}'
            )
        else:
            curl = cfields.curl
        chapter = Chapter(self, ch_link.link_text, curl, ch_link.link_title, previous=prev_chapter)
        if prev_chapter is not None:
            prev_chapter.__set_next__(chapter)
        self.__chapter_cpath_tuples.append((chapter, content_cpath))
        self.__cpath_to_chapter[content_cpath] = chapter

        return chapter

    def get_chapter(self, cpath, default=None):
        return self.__cpath_to_chapter.get(cpath, default)

    def __on_mutation_done__(self):
        assert not self.__is_done

        # delete temporaries
        del self.__chapter_cpath_tuples

        self.__is_done = True

    def get_cpath_chapter_map(self):
        return self.__cpath_to_chapter.copy()

    @property
    def parts(self):
        return tuple(self.__chapter_parts)

    def get_last_part(self):
        if self.__chapter_parts:
            return self.__chapter_parts[-1]
        else:
            return None

    def as_html(self, added_level=None):
        added_level = 0 if added_level is None else added_level
        res = []
        for ch_part in self.__chapter_parts:
            res.append(
                ch_part.as_html(added_level=added_level)
            )
        return '\n'.join(res)

    def __str__(self):
        return '\n'.join(str(part) for part in self.__chapter_parts)

    def __repr__(self):
        return repr(self.__str__())


if __name__ == '__main__':
    text = \
"""
* chapter 1
    - [Sub Chapter 1](/go1)
    - Subchapter 2 whithout link
        * sub under sub 2
        * another sub under sub 2
- [Chapter 2](/ch2)
- Ch 1 Pt 2

1. ch p none
    - sub ch p none
        - more sub ch p none
- And this

"""
    p = MarkedChapterParser(None, text)
    toc = p.parse()
    print()
    # for p in toc.parts:
    #     print(p)
    print(toc.as_html(added_level=1))



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
from synamic.core.standalones.functions.decorators import not_loaded
from synamic.exceptions import SynamicSydParseError, SynamicErrors, SynamicFSError


class _Menu:
    def __init__(self, site, title=None, link=None, id=None, children=None, other_fields=None):
        assert title is not None
        self.__site = site
        self.__title = title
        self.__link = link
        self.__id = id
        self.__children = children if children is not None else ()
        self.__other_fields = {} if other_fields is None else other_fields

    def __iter__(self):
        return iter(self.__children)

    @property
    def title(self):
        return self.__title

    @property
    def link(self):
        return self.__site.object_manager.getc(self.__link)

    @property
    def children(self):
        return self.__children

    def __getattr__(self, item):
        return self.__other_fields.get(item, None)

    def __getitem__(self, item):
        return self.__other_fields.get(item, None)

    def __str__(self):
        return self.title + " (%s) : " % self.link + str(self.children)

    def __repr__(self):
        return repr(str(self))


class MenuService:
    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False
        self.__menu_dir = self.__site.system_settings['dirs.metas.menus']

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        # for file_path in file_paths:
        self.__is_loaded = True

    def make_menu(self, name):
        path_tree = self.__site.get_service('path_tree')
        fn = name + '.syd'
        path = path_tree.create_file_cpath(self.__menu_dir, fn)
        try:
            syd = self.__site.object_manager.get_syd(path)
        except (SynamicSydParseError, SynamicFSError) as e:
            raise SynamicErrors(
                f"Synamic error during making menu object. Error occurred for menu named {name}",
                e
            )

        menu_obj = _Menu(
            self.__site,
            title=name,
            link='__root_link__',
            children=self._process_menu(syd)
        )
        return menu_obj

    def get_menu_names(self):
        menu_names = []
        path_tree = self.__site.get_service('path_tree')
        menu_cdir = path_tree.create_dir_cpath(self.__menu_dir)
        if menu_cdir.exists():
            menu_cfiles = menu_cdir.list_files(depth=1)
            for menu_cfile in menu_cfiles:
                basename = menu_cfile.basename
                if basename.lower().endswith('.syd'):
                    menu_name = basename[:-len('.syd')]
                    menu_names.append(menu_name)
        return tuple(menu_names)

    def _process_menu(self, starting_menu_syd) -> tuple:
        res_list = []
        menus = starting_menu_syd.get('menus', None)
        if menus is not None:
            for menu in menus:
                dict_flat = menu
                title = dict_flat.get('title', None)
                assert title is not None
                del dict_flat['title']
                link = dict_flat.get('link', None)
                if link is not None:
                    del dict_flat['link']
                res_list.append(
                    _Menu(
                        self.__site,
                        title=title,
                        link=link,
                        children=self._process_menu(menu),
                        other_fields=dict_flat
                    )
                )
        return tuple(res_list)




from synamic.core.object_manager.query import SimpleQueryParser
from synamic.core.standalones.functions.date_time import parse_datetime, parse_date, parse_time
from .types import Html, Markdown
import re
import datetime
from synamic.exceptions import SynamicMarkNotFound
from synamic.exceptions import SynamicInvalidNumberFormat


class _Pat:
    newline_pat = re.compile(r'\r\n|\n|\r', re.MULTILINE)
    number_pat = re.compile(r'''
                            ^(?P<number>
                                (?P<sign>[-+])?
                                (?P<int>[0-9]+)
                                (\.
                                    (?P<decimal>[0-9]+)
                                )?
                            )$''', re.VERBOSE)

    separator_comma_pat = re.compile(r',[^,]?')
    # separator_comma_pat = re.compile(r',[^,]*')
    multiple_commas_pat = re.compile(r'(?P<start>[^,]*)(?P<commas>,{2,})(?P<end>[^,]*)')
    # multiple_commas_pat = re.compile(r'(?P<start>[^,]*)(?P<commas>,{2})*(?P<end>[^,]*)')

    @staticmethod
    def multiple_commas_repl_fun(mobj):
        commas = mobj.group('commas')
        commas = commas[1:]
        start = mobj.group('start')
        end = mobj.group('end')
        if start is None:
            start = ''
        if end is None:
            end = ''
        return start + commas + end


_default_types = frozenset({
        'number',
        'string',
        'text',
        'date',
        'time',
        'datetime',
        'markdown',
        'html',

        'marker#type',
        'marker#tags',
        'marker#categories',
        'user',

        # TODO: add the following __document_types
        # template
        # image
        # attachment
        # etc

        'number[]',
        'string[]',
        'date[]',
        'time[]',
        'datetime[]',
})

_class_map = {}


def _add_converter_type(name):
    assert name in _default_types

    def class_or_fun_decorator(cls_or_fun):
        assert name not in _class_map
        _class_map[name] = cls_or_fun
        return cls_or_fun
    return class_or_fun_decorator


class ConverterCallable:
    def __init__(self, type_system, name, supported_compare_ops=frozenset()):
        self.__type_system = type_system
        self.__name = name
        self.__supported_compare_ops = frozenset(supported_compare_ops)
        # validation
        for op in self.__supported_compare_ops:
            assert op in SimpleQueryParser.COMPARE_OPERATORS_SET

    @property
    def type_system(self):
        return self.__type_system

    @property
    def name(self):
        return self.__name

    @property
    def supported_compare_ops(self):
        return self.__supported_compare_ops

    def supports_compare_op(self, op):
        return op in self.supported_compare_ops

    @classmethod
    def validate_op_value(cls, op, left_value, right_value):
        assert op in SimpleQueryParser.COMPARE_OPERATORS_SET
        if op in ('in', '!in'):
            assert isinstance(right_value, (list, tuple))
            assert not isinstance(left_value, (list, tuple))
        elif op in ('contains', '!contains'):
            assert isinstance(left_value, (list, tuple))
            assert not isinstance(right_value, (list, tuple))
        else:
            assert not isinstance(left_value, (list, tuple))
            assert not isinstance(right_value, (list, tuple))

    def compare(self, op, left_value, right_value):
        assert op in self.supported_compare_ops, f'op: {op}, supported ops: {self.supported_compare_ops}'
        self.validate_op_value(op, left_value, right_value)
        return SimpleQueryParser.COMPARE_OPERATOR_2_PY_OPERATOR_FUN[op](left_value, right_value)

    def __call__(self, value, *args, **kwargs):
        raise NotImplemented


class ConverterCallableListCompareMixin(ConverterCallable):
    def compare(self, op, left_value, right_value):
        assert op in self.supported_compare_ops
        if op in ('contains', '!contains'):
            right_value = right_value[0]
            self.validate_op_value(op, left_value, right_value)
            return SimpleQueryParser.COMPARE_OPERATOR_2_PY_OPERATOR_FUN[op](left_value, right_value)
        else:
            assert op in ('in', '!in')
            assert isinstance(left_value, (list, tuple)), '%s : %s' % (str(type(left_value)), str(left_value))
            assert isinstance(right_value, (list, tuple)), '%s : %s' % (str(type(right_value)), str(right_value))
            for single_value in left_value:
                if single_value not in right_value:
                    return False
            return True


@_add_converter_type('number')
class NumberConverter(ConverterCallable):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            '>',
            '<',
            '>=',
            '<=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        if isinstance(txt, (int, float)):
            return txt
        m_number = _Pat.number_pat.match(txt)
        if not m_number:
            raise SynamicInvalidNumberFormat(f"Invalid number format: {txt}")
        else:
            sign_str = m_number.group('sign')
            int_str = m_number.group('int')
            decimal_str = m_number.group('decimal')
            int_str = int_str.lstrip('0')
            if int_str == '':
                int_str = '0'
            num_str = sign_str + int_str + decimal_str

            if decimal_str:
                num = float(num_str)
            else:
                num = int(num_str)
            return num


@_add_converter_type('date')
class DateConverter(ConverterCallable):
    """May include locale in future"""
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            '>',
            '<',
            '>=',
            '<=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        if isinstance(txt, datetime.date):
            return txt
        return parse_date(txt)


@_add_converter_type('time')
class TimeConverter(ConverterCallable):
    """May include locale in future"""
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            '>',
            '<',
            '>=',
            '<=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        if isinstance(txt, datetime.time):
            return txt
        return parse_time(txt)


@_add_converter_type('datetime')
class DateTimeConverter(ConverterCallable):
    """May include locale in future"""
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            '>',
            '<',
            '>=',
            '<=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        if isinstance(txt, datetime.datetime):
            return txt
        return parse_datetime(txt)


@_add_converter_type('string')
class StringConverter(ConverterCallable):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        """Strings are single line things, so, any multi-line will be skipped and the first line will be taken"""
        if isinstance(txt, (int, float)):
            txt = str(txt)
        assert isinstance(txt, str)
        txts = _Pat.newline_pat.split(txt)
        return txts[0]


@_add_converter_type('text')
class TextConverter(ConverterCallable):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        """return as is"""
        assert isinstance(txt, str)
        return txt


@_add_converter_type('markdown')
class MarkdownConverter(ConverterCallable):
    def __call__(self, txt, value_pack=None, *args, **kwargs):
        """
        Return a markdown instance.
        """
        return Markdown(self.type_system.site, txt, value_pack=value_pack)


@_add_converter_type('html')
class HtmlConverter(ConverterCallable):
    def __call__(self, txt, *args, **kwargs):
        """
        Probably should return a Html instance.
        """
        return Html(txt)


@_add_converter_type('number[]')
class NumberListConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        number_converter = self.type_system.get_converter('number')
        txts = txt.split(',')
        num_strs = [x.strip() for x in txts]
        numbers = []
        for num_str in num_strs:
            number = number_converter(num_str)
            numbers.append(number)
        return tuple(numbers)


@_add_converter_type('string[]')
class StringListConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        """
        Separator for string is comma (,) - so what if someone want to put a comma inside their string?
         Ans: precede that with another comma.
        So, when more than consecutive comma is found the first one is skipped and the later ones are considered as literal commas
        """
        res = []
        strings = _Pat.separator_comma_pat.split(txt)
        for string in strings:
            string = string.strip()
            string = _Pat.multiple_commas_pat.sub(_Pat.multiple_commas_repl_fun, string)
            res.append(string)
        return tuple(strings)


@_add_converter_type('date[]')
class DateListConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        date_converter = self.type_system.get_converter('date')
        res = []
        txt = txt.strip()
        date_str_list = txt.split(',')
        for date_str in date_str_list:
            date_str = date_str.strip()
            res.append(
                date_converter(date_str)
            )
        return tuple(res)


@_add_converter_type('time[]')
class TimeListConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        time_converter = self.type_system.get_converter('time')
        res = []
        txt = txt.strip()
        time_str_list = txt.split(',')
        for time_str in time_str_list:
            time_str = time_str.strip()
            res.append(
                time_converter(time_str)
            )
        return tuple(res)


@_add_converter_type('datetime[]')
class DateTimeListConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        datetime_converter = self.type_system.get_converter('datetime')
        res = []
        txt = txt.strip()
        datetime_str_list = txt.split(',')
        for datetime_str in datetime_str_list:
            datetime_str = datetime_str.strip()
            res.append(
                datetime_converter(datetime_str)
            )
        return tuple(res)


@_add_converter_type('marker#type')
class MarkTypeConverter(ConverterCallable):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        object_manager = self.type_system.site.object_manager
        type_marker = object_manager.get_marker('type')
        title = txt.strip()
        mark = type_marker.get_mark_by_title(title, None)
        if mark is None:
            mark = type_marker.make_mark_by_title(txt)
            # TODO: is the txt always title? not ID? I gotta fix this.
            # raise SynamicMarkNotFound(f'Mark with title {title} does not exist')
        return mark


@_add_converter_type('marker#tags')
class MarkTagsConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        """Strings are single line things, so, any multi-line will be skipped and the first line will be taken"""
        object_manager = self.type_system.site.object_manager
        tag_marker = object_manager.get_marker('tags')
        res = []
        tag_titles = _Pat.separator_comma_pat.split(txt)
        for title in tag_titles:
            title = title.strip()
            if title != '':
                title = _Pat.multiple_commas_pat.sub(_Pat.multiple_commas_repl_fun, title)
                mark = tag_marker.get_mark_by_title(title, None)
                if mark is None:
                    mark = tag_marker.make_mark_by_title(title)
                res.append(
                    mark
                )
        return tuple(res)


@_add_converter_type('marker#categories')
class MarkCategoriesConverter(ConverterCallableListCompareMixin):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            'contains',
            '!contains',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, txt, *args, **kwargs):
        object_manager = self.type_system.site.object_manager
        categories_marker = object_manager.get_marker('categories')
        res = []
        categories_titles = _Pat.separator_comma_pat.split(txt)
        for title in categories_titles:
            title = title.strip()
            if title != '':
                title = _Pat.multiple_commas_pat.sub(_Pat.multiple_commas_repl_fun, title)
                mark = categories_marker.get_mark_by_title(title, None)
                if mark is None:
                    mark = categories_marker.make_mark_by_title(title)
                res.append(
                    mark
                )
        return tuple(res)


# TODO: implement in !in for all the converter that returns single value - currently it does not work.
@_add_converter_type('user')
class UserConverter(ConverterCallable):
    def __init__(self, type_system, name):
        supported_ops = frozenset([
            '==',
            '!=',
            'in',
            '!in'
        ])
        super().__init__(type_system, name, supported_ops)

    def __call__(self, user_id, *args, **kwargs):
        object_manager = self.type_system.site.object_manager
        user = object_manager.get_user(user_id)
        return user
